# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
@software{jax2018github,
  author = {James Bradbury and Roy Frostig and Peter Hawkins and Matthew James Johnson and Chris Leary and Dougal Maclaurin and George Necula and Adam Paszke and Jake Vander{P}las and Skye Wanderman-{M}ilne and Qiao Zhang},
  title = {{JAX}: composable transformations of {P}ython+{N}um{P}y programs},
  url = {http://github.com/google/jax},
  version = {0.2.5},
  year = {2018},
}
Datasets used in examples.
"""


import array
import gzip
import os
from os import path
import struct
import urllib.request
import cv2
import numpy as np

import numpy.random as npr

_DATA = "/opt/ml/tmp/jax_example_data/"


def center_data(X):
    mean_x = np.mean(X, axis=0, keepdims=True)
    reduced_mean = np.subtract(X, mean_x)
    reduced_mean = reduced_mean.astype(np.float32)
    return reduced_mean


def synth_batches(input_shape):
    while True:
        images = np.random.rand(*input_shape).astype("float32")
        yield images


def _download(url, filename):
    """Download a url to a file in the JAX data temp directory."""
    if not path.exists(_DATA):
        os.makedirs(_DATA)
    out_file = path.join(_DATA, filename)
    if not path.isfile(out_file):
        urllib.request.urlretrieve(url, out_file)
        print("downloaded {} to {}".format(url, _DATA))


def _partial_flatten(x):
    """Flatten all but the first dimension of an ndarray."""
    return np.reshape(x, (x.shape[0], -1))


def _one_hot(x, k, dtype=np.float32):
    """Create a one-hot encoding of x of size k."""
    return np.array(x[:, None] == np.arange(k), dtype)


def mnist_raw():
    """Download and parse the raw MNIST dataset."""
    # CVDF mirror of http://yann.lecun.com/exdb/mnist/
    base_url = "https://storage.googleapis.com/cvdf-datasets/mnist/"

    def parse_labels(filename):
        with gzip.open(filename, "rb") as fh:
            _ = struct.unpack(">II", fh.read(8))
            return np.array(array.array("B", fh.read()), dtype=np.uint8)

    def parse_images(filename):
        with gzip.open(filename, "rb") as fh:
            _, num_data, rows, cols = struct.unpack(">IIII", fh.read(16))
            return np.array(array.array("B", fh.read()), dtype=np.uint8).reshape(
                num_data, rows, cols
            )

    for filename in [
        "train-images-idx3-ubyte.gz",
        "train-labels-idx1-ubyte.gz",
        "t10k-images-idx3-ubyte.gz",
        "t10k-labels-idx1-ubyte.gz",
    ]:
        _download(base_url + filename, filename)

    train_images = parse_images(path.join(_DATA, "train-images-idx3-ubyte.gz"))
    train_labels = parse_labels(path.join(_DATA, "train-labels-idx1-ubyte.gz"))
    test_images = parse_images(path.join(_DATA, "t10k-images-idx3-ubyte.gz"))
    test_labels = parse_labels(path.join(_DATA, "t10k-labels-idx1-ubyte.gz"))

    return train_images, train_labels, test_images, test_labels


def img_resize(train_images):
    train_data = []
    for img in train_images:
        resized_img = cv2.resize(img, (6, 6))
        train_data.append(resized_img)
    return np.asarray(train_data)

def img_resize_noise(train_images):
    train_data = []
    for img in train_images:
        #resized_img = cv2.resize(img, (6, 6))
        dst = cv2.GaussianBlur(img, (5, 5), 0.0, cv2.BORDER_DEFAULT)
        #train_data.append(resized_img)
    return dst #np.asarray(train_data)


def mnist(permute_train=False, resize=False, filter=False):
    """Download, parse and process MNIST data to unit scale and one-hot labels."""

    train_images, train_labels, test_images, test_labels = mnist_raw()
    if filter:
        train_filter = np.where(train_labels == 1)
        train_images = train_images[train_filter]
        train_labels = train_labels[train_filter]

        test_filter = np.where(test_labels == 1)
        test_images = test_images[test_filter]
        test_labels = test_labels[test_filter]

    if resize:
        train_images = img_resize(train_images)
        test_images = img_resize(test_images)
    train_images = _partial_flatten(train_images) / np.float32(255.0)
    test_images = _partial_flatten(test_images) / np.float32(255.0)
    train_labels = _one_hot(train_labels, 10)
    test_labels = _one_hot(test_labels, 10)
    if permute_train:
        perm = np.random.RandomState(0).permutation(train_images.shape[0])
        train_images = train_images[perm]
        train_labels = train_labels[perm]
    return train_images, train_labels, test_images, test_labels


def meta_mnist(batch_size, filter=False):
    train_len = 60000
    test_len = 10000
    if filter:
        train_len= 6000
        test_len = 1000
    num_complete_batches, leftover = divmod(train_len, batch_size)
    num_batches = num_complete_batches + bool(leftover)
    return locals()


def get_mnist_data(batch_size, resize, filter=False):
    train_images, train_labels, test_images, test_labels = mnist_raw()
    if resize:
        train_images = img_resize(train_images)
        #test_images = img_resize(test_images)
    train_images = _partial_flatten(train_images) / np.float32(255.0)
    #test_images = _partial_flatten(test_images) / np.float32(255.0)
    if filter:
        train_filter = np.where(train_labels == 1)
        train_images = train_images[train_filter]
        train_labels = train_labels[train_filter]
    train_labels = _one_hot(train_labels, 10)
    #test_labels = _one_hot(test_labels, 10)
    train_images = center_data(train_images)
    total_data_len = train_images.shape[0]
    num_complete_batches, leftover = divmod(total_data_len, batch_size)
    num_batches = num_complete_batches + bool(leftover)

    rng = npr.RandomState(0)
    while True:
        perm = rng.permutation(total_data_len)
        for i in range(num_batches):
            batch_idx = perm[i * batch_size : (i + 1) * batch_size]
            yield train_images[batch_idx], train_labels[batch_idx]


def get_preload_mnist_data(train_images, train_labels, test_images, test_labels, batch_size, resize, filter=False):
    if resize:
        train_images = img_resize(train_images)
        #test_images = img_resize(test_images)
    train_images = _partial_flatten(train_images) / np.float32(255.0)
    #test_images = _partial_flatten(test_images) / np.float32(255.0)
    if filter:
        train_filter = np.where(train_labels == 1)
        train_images = train_images[train_filter]
        train_labels = train_labels[train_filter]
    train_labels = _one_hot(train_labels, 10)
    #test_labels = _one_hot(test_labels, 10)
    train_images = center_data(train_images)
    total_data_len = train_images.shape[0]
    num_complete_batches, leftover = divmod(total_data_len, batch_size)
    num_batches = num_complete_batches + bool(leftover)

    rng = npr.RandomState(0)
    while True:
        perm = rng.permutation(total_data_len)
        for i in range(num_batches):
            batch_idx = perm[i * batch_size : (i + 1) * batch_size]
            yield train_images[batch_idx], train_labels[batch_idx]


def mnist_preprocess_cont(resize, filter, center=True):
    train_images, train_labels, test_images, test_labels = mnist(
        permute_train=False, resize=resize, filter=filter)
    if center:
        train_images = center_data(train_images)
        test_images = center_data(test_images)
    return train_images, train_labels, test_images, test_labels


def get_mnist_batch_alter(train_images, train_labels, test_images,
                    test_labels, alter: list, batch_size, resize, filter=False):
    alter = 1.0 - alter[0]
    if alter<0.1:
        alter = 0.1
    # alter batch len
    batch_size = int(batch_size/alter)
    print("batch_size: ",batch_size)
    total_data_len = train_images.shape[0]
    num_complete_batches, leftover = divmod(total_data_len, batch_size)
    num_batches = num_complete_batches + bool(leftover)
    rng = npr.RandomState(0)
    while True:
        perm = rng.permutation(total_data_len)
        for i in range(num_batches):
            batch_idx = perm[i * batch_size : (i + 1) * batch_size]
            yield train_images[batch_idx], train_labels[batch_idx]



if __name__ == "__main__":
    train_images, _, _, _ = mnist_raw()
    #train_images = _partial_flatten(train_images) / np.float32(255.0)
    for src in train_images:
        #src = train_images[0]#/np.float32(255.0)
        #print(src)
        #print(src/np.float32(255.0))
        #print(src.shape)
        dst = adjust_gamma(src, 0.5)
        #dst = cv2.medianBlur(src, 3)
        #src =cv2.resize(src, (6,6))
        #dst = cv2.resize(dst, (6, 6))
        print(dst.shape)
    # cv2.imshow("Gaussian Smoothing", np.hstack((src, dst)))
    # cv2.waitKey(0)  # waits until a key is pressed
    # cv2.destroyAllWindows()
    # print(train_images.shape[0])
    # print(test_images.shape[0])
    # # z = meta_mnist(5000)
    # print(meta_mnist(6742, True)["num_batches"])



import jax.numpy as jnp
from jax import grad, hessian, jit, random
from jax.flatten_util import ravel_pytree
from jax.experimental.optimizers import l2_norm
from jax.nn import softmax
from jax.nn import log_softmax, logsumexp
key = random.PRNGKey(0)
def sigmoid(x):
    return 0.5 * (jnp.tanh(x / 2) + 1)

# Outputs probability of a label being true.
def predict(W, b, inputs):
    return jnp.dot(inputs, W) + b

# Build a toy dataset.
inputs = jnp.array([[0.52, 1.12,  0.77],
                   [0.88, -1.08, 0.15],
                   [0.52, 0.06, -1.30],
                   [0.74, -2.49, 1.39]])
targets = jnp.array([1, 1, 2, 3])
def _one_hot(x, k, dtype=jnp.float32):
    """Create a one-hot encoding of x of size k."""
    return jnp.array(x[:, None] == jnp.arange(k), dtype)
targets = _one_hot(targets, 3)


def loss(W, b):
    logits = predict(W, b, inputs)
    preds = logits - logsumexp(logits, axis=1, keepdims=True)
    loss = -jnp.mean(jnp.sum(preds * targets, axis=1))
    loss += 0.001* (l2_norm(W) + l2_norm(b))
    return loss



# Initialize random model coefficients
key, W_key, b_key = random.split(key, 3)
W = random.normal(W_key, (3,3))
b = random.normal(b_key, (3,))

f = lambda W: predict(W, b, inputs)
print("loss:",loss(W, b))
for i in range(1000):
    W_grad, b_grad = grad(loss, (0, 1))(W, b)
    W = W -0.005*W_grad
    b = b - 0.005 * b_grad
print("loss:",loss(W, b))
H = hessian(loss)(W, b)
h, _ = ravel_pytree(H)
eigen_vals = jnp.linalg.eigvals(h.reshape(9,9)).real
eigen_vals = sorted(eigen_vals, reverse=True)
print(eigen_vals) # should be all positive for convex function
#outputs: [0.2406996, 0.16962789, 0.13137847, 0.07919562, 0.037625454, 0.02834966, 0.00042202187, 0.00042201488, 0.00037239227]


import jax.numpy as np
import numpy as onp
import numpy.random as npr
from cjax.utils.abstract_problem import AbstractProblem
from jax.tree_util import tree_map

import jax.numpy as jnp
from jax import random
from jax.experimental import stax
from jax.experimental.stax import Dense, FanOut, Relu, Softplus


batch_size = 8
input_shape = (batch_size, 8)
step_size = 0.1
num_steps = 10
code_dim = 1
npr.seed(7)


def synth_batches():
    while True:
        images = npr.rand(*input_shape).astype("float32")
        yield images


batches = synth_batches()
inputs = next(batches)
u, s, v_t = onp.linalg.svd(inputs, full_matrices=False)
I = np.eye(v_t.shape[-1])
I_add = npr.normal(0.0, 0.002, size=I.shape)
noisy_I = I + I_add

encoder_init, encode = stax.serial(
    Dense(512),
    Relu,
    Dense(512),
    Relu,
    FanOut(2),
    stax.parallel(Dense(10), stax.serial(Dense(10), Softplus)),
)

decoder_init, decode = stax.serial(
    Dense(512),
    Relu,
    Dense(512),
    Relu,
    Dense(8),
)


def gaussian_kl(mu, sigmasq):
    """KL divergence from a diagonal Gaussian to the standard Gaussian."""
    return -0.5 * jnp.sum(1.0 + jnp.log(sigmasq) - mu ** 2.0 - sigmasq)


def gaussian_sample(rng, mu, sigmasq):
    """Sample a diagonal Gaussian."""
    return mu + jnp.sqrt(sigmasq) * random.normal(rng, mu.shape)


def bernoulli_logpdf(logits, x):
    """Bernoulli log pdf of data x given logits."""
    return -jnp.sum(jnp.logaddexp(0.0, jnp.where(x, -1.0, 1.0) * logits))


def elbo(rng, params, bparam, images):
    """Monte Carlo estimate of the negative evidence lower bound."""
    enc_params, dec_params = params
    mu_z, sigmasq_z = encode(enc_params, images)
    logits_x = decode(dec_params, gaussian_sample(rng, mu_z, sigmasq_z))
    return bernoulli_logpdf(logits_x, images) - bparam[0] * gaussian_kl(mu_z, sigmasq_z)


class TopologyVAE(AbstractProblem):
    def __init__(self):
        self.HPARAMS_PATH = "examples/vae/hparams.json"

    @staticmethod
    def objective(params, bparam) -> float:
        elbo_rng, data_rng = random.split(random.PRNGKey(1))
        loss = -elbo(elbo_rng, params, bparam, inputs)
        # logits = predict_fun(params, inputs, bparam=bparam[0], activation_func=sigmoid)
        # loss = np.mean(np.square((np.subtract(logits, inputs))))
        # loss += 0.1*(l2_norm(params) + l2_norm(bparam))
        return np.mean(loss)

    def initial_value(self):
        enc_init_rng, dec_init_rng = random.split(random.PRNGKey(2))
        _, init_encoder_params = encoder_init(enc_init_rng, (batch_size, 8))  # 28*28
        _, init_decoder_params = decoder_init(dec_init_rng, (batch_size, 10))
        init_params = [init_encoder_params, init_decoder_params]
        bparam = [np.array([0.01], dtype=np.float64)]
        return init_params, bparam

    def initial_values(self):
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.005, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams


if __name__ == "__main__":
    problem = TopologyVAE()
    ae_params, bparams = problem.initial_value()
    loss = problem.objective(ae_params, bparams)
    print(loss)

    # init_c = constant_2d(I)
    # print(init_c(key=0, shape=(8,8)))



from cjax.utils.abstract_problem import AbstractProblem
import jax.numpy as np
from jax import vmap
from jax import random
from jax.scipy.special import logsumexp
from jax.tree_util import *

inputs = random.normal(random.PRNGKey(1), (1, 2 * 2))
outputs = np.ones(shape=(1, 1))


class SimpleNeuralNetwork(AbstractProblem):
    def __init__(self):
        pass

    @staticmethod
    def objective(params, bparam) -> float:
        def neural_net_predict(params, bparam, inputs):
            # per-example predictions
            activations = inputs
            for w, b in params[:-1]:
                activations = np.dot(w, activations) + b + bparam
                # activations = relu(_outputs)

            final_w, final_b = params[-1]
            logits = np.dot(final_w, activations) + final_b
            return logits - logsumexp(logits)

        def regularizer_l2(params):
            reg_add = 0.0
            for (w, b) in params:
                reg_add += np.sum([np.sum(np.square(w)), np.sum(np.square(b))])
            return reg_add

        # vectorization of mini-batch of data.
        # #3rd argument's 0th-axis is vmaped. --> inputs(10)
        batched_predict = vmap(neural_net_predict, in_axes=(None, None, 0))

        preds = batched_predict(params, bparam, inputs)
        mse = -np.mean(preds * outputs)
        reg = regularizer_l2(params)
        return mse + reg

    def init_network_params(self, sizes, key):
        keys = random.split(key, len(sizes))
        return [
            self.random_layer_params(m, n, k)
            for m, n, k in zip(sizes[:-1], sizes[1:], keys)
        ]

    @staticmethod
    def random_layer_params(m, n, key, scale=1e-1):
        w_key, b_key = random.split(key)
        return scale * random.normal(w_key, (n, m)), scale * random.normal(b_key, (n,))

    def initial_value(self) -> Tuple:
        bparam = [np.array([-0.50], dtype=np.float64)]
        layer_sizes = [4, 1]
        state = self.init_network_params(layer_sizes, random.PRNGKey(0))
        return state, bparam

    def initial_values(self) -> Tuple:
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.05, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams



from abc import ABC, abstractmethod

from jax.experimental.optimizers import adam, sgd, nesterov, adagrad, momentum, rmsprop
from jax.experimental.optimizers import optimizer, make_schedule


@optimizer
def gradient_ascent(step_size):
    """Construct optimizer triple for stochastic gradient descent.

    Args:
      step_size: positive scalar, or a callable representing a step size schedule
        that maps the iteration index to positive scalar.

    Returns:
      An (init_fun, update_fun, get_params) triple.
    """
    step_size = make_schedule(step_size)

    def init(x0):
        return x0

    def update(i, g, x):
        return x + step_size(i) * g

    def get_params(x):
        return x

    return init, update, get_params


class Optimizer:
    """Abstract Optimizer to be inherited by developer for any new optimizer."""

    def __init__(self, lr):
        self._lr = lr

    @property
    def lr(self):
        return self._lr

    @lr.setter
    def lr(self, lr):
        self._lr = lr

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        pass


class GDOptimizer(Optimizer):
    def __init__(self, learning_rate):
        super().__init__(learning_rate)
        self.opt_init, self.opt_update, self.get_params = sgd(step_size=self.lr)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params


class _GDOptimizer(Optimizer):
    def __init__(self, learning_rate):
        super().__init__(learning_rate)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        for (w, dw) in zip(params, grad_params):
            print(w, dw)
        return [w - self.lr * dw for (w, dw) in zip(params, grad_params)]


class GAOptimizer(Optimizer):
    def __init__(self, learning_rate):
        super().__init__(learning_rate)
        self.opt_init, self.opt_update, self.get_params = gradient_ascent(
            step_size=self.lr
        )

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params


class AdamOptimizer(Optimizer):
    def __init__(self, learning_rate):
        super().__init__(learning_rate)
        self.opt_init, self.opt_update, self.get_params = adam(step_size=self.lr)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params



class MomentumOptimizer(Optimizer):
    def __init__(self, learning_rate, mass=0.9):
        super().__init__(learning_rate)
        self.mass = mass
        self.opt_init, self.opt_update, self.get_params = momentum(step_size=self.lr, mass=self.mass)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params


class NestrovOptimizer(Optimizer):
    def __init__(self, learning_rate, mass=0.9):
        super().__init__(learning_rate)
        self.mass = mass
        self.opt_init, self.opt_update, self.get_params = nesterov(step_size=self.lr, mass=self.mass)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params

class RMSPropOptimizer(Optimizer):
    def __init__(self, learning_rate):
        super().__init__(learning_rate)
        self.opt_init, self.opt_update, self.get_params = rmsprop(step_size=self.lr)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params

class AdagradOptimizer(Optimizer):
    def __init__(self, learning_rate):
        super().__init__(learning_rate)
        self.opt_init, self.opt_update, self.get_params = adagrad(step_size=self.lr)

    def update_params(
        self, params: list, grad_params: list, step_index: int = 0
    ) -> list:
        opt_state = self.opt_init(params)
        params = self.get_params(self.opt_update(step_index, grad_params, opt_state))
        return params


class OptimizerCreator:
    def __init__(self, opt_string, learning_rate):
        self.learning_rate = learning_rate
        self._opt_string = opt_string

    def get_optimizer(self) -> Optimizer:
        if self._opt_string == "gradient-descent":
            return GDOptimizer(self.learning_rate)
        elif self._opt_string == "adam":
            return AdamOptimizer(self.learning_rate)
        elif self._opt_string == "adagrad":
            return AdagradOptimizer(self.learning_rate)
        elif self._opt_string == "rmsprop":
            return RMSPropOptimizer(self.learning_rate)
        elif self._opt_string == "nestrov":
            return NestrovOptimizer(self.learning_rate)
        elif self._opt_string == "adam":
            return AdamOptimizer(self.learning_rate)
        elif self._opt_string == "gradient-ascent":
            return GAOptimizer(self.learning_rate)
        else:
            print(f"Optimizer not implemented: {self._opt_string} using default gradient-descent")
            return GDOptimizer(self.learning_rate)


if __name__ == "__main__":
    opt = GDOptimizer(0.5)
    print(opt.lr)



from cjax.utils.abstract_problem import ProblemWraper
from typing import Dict
from cjax.continuation.base_continuation import Continuation
from cjax.continuation.natural_continuation import NaturalContinuation
from cjax.continuation.secant_continuation import SecantContinuation
from cjax.continuation._arc_len_continuation import PseudoArcLenContinuation
from cjax.continuation._perturbed_arc_len_continuation import (
    PerturbedPseudoArcLenContinuation,
)
from cjax.continuation.perturbed_fixed_arclen_continuation import (
    PerturbedPseudoArcLenFixedContinuation,
)


class ContinuationCreator:
    """Continuation Factory to create the right objects on the fly.
    TODO: Use **kwargs to reduce the size of the constructors.
    """

    def __init__(self, problem: ProblemWraper, hparams: Dict, key=0, mlflow=None):
        self.problem = problem
        self.hparams = hparams
        self.key = key
        self.mlflow = mlflow

    def get_continuation_method(self) -> Continuation:
        """Creates the continuation object based on user arguments.

        options-
        natural: monotonic update of continuation parameter, followed by unconstrained solver
        secant: approximates tangent of the solution path to update both state and continuation parameter,
                followed by unconstrained solver
        parc: To be deprecated (TODO)
        parc-perturb: secant predictor, Langrange min-max updates for corrector with orthogonal constraint
        parc-fix-perturb: secant predictor, Langrange multiplier is fixed (hyperparamter)
                          updates for corrector with orthogonal constraint

        Returns:
            object: Continuation
        Raises:
            NotImplementedError
        """
        if self.hparams["meta"]["method"] == "natural":
            state, bparam = self.problem.initial_value()
            return NaturalContinuation(
                state,
                bparam,
                counter=0,
                objective=self.problem.objective,
                accuracy_fn=self.problem.accuracy_fn,
                hparams=self.hparams,
            )
        elif self.hparams["meta"]["method"] == "secant":
            states, bparams = self.problem.initial_values()
            state, bparam = states[1], bparams[1]
            state_0, bparam_0 = states[0], bparams[0]
            return SecantContinuation(
                state,
                bparam,
                state_0,
                bparam_0,
                counter=0,
                objective=self.problem.objective,
                accuracy_fn=self.problem.accuracy_fn,
                hparams=self.hparams,
            )
        elif self.hparams["meta"]["method"] == "parc":
            states, bparams = self.problem.initial_values()
            state, bparam = states[1], bparams[1]
            state_0, bparam_0 = states[0], bparams[0]
            return PseudoArcLenContinuation(
                state,
                bparam,
                state_0,
                bparam_0,
                counter=0,
                objective=self.problem.objective,
                dual_objective=self.problem.dual_objective,
                hparams=self.hparams,
            )
        elif self.hparams["meta"]["method"] == "parc-perturb":
            states, bparams = self.problem.initial_values()
            state, bparam = (
                states[1],
                bparams[1],
            )  # TODO: remove this hard-coding, basically use the previous two solutions.
            state_0, bparam_0 = states[0], bparams[0]
            return PerturbedPseudoArcLenContinuation(
                state,
                bparam,
                state_0,
                bparam_0,
                counter=0,
                objective=self.problem.objective,
                dual_objective=self.problem.dual_objective,
                hparams=self.hparams,
                key_state=self.key,
            )
        elif self.hparams["meta"]["method"] == "parc-fix-perturb":
            states, bparams = self.problem.initial_values()
            state, bparam = states[1], bparams[1]
            state_0, bparam_0 = states[0], bparams[0]
            return PerturbedPseudoArcLenFixedContinuation(
                state,
                bparam,
                state_0,
                bparam_0,
                counter=0,
                objective=self.problem.objective,
                dual_objective=self.problem.dual_objective,
                accuracy_fn = self.problem.accuracy_fn,
                hparams=self.hparams,
                key_state=self.key,
            )
        else:
            raise NotImplementedError



from cjax.utils.abstract_problem import AbstractProblem
import jax.numpy as np
from jax import random
from jax.experimental.optimizers import l2_norm
from jax.tree_util import *
from flax import linen as nn  # The Linen API
import jax
import numpy.random as npr
from cjax.utils import datasets

#
# num_classes = 10
# inputs = random.normal(random.PRNGKey(1), (1, 10, 10))
# outputs = np.ones(shape=(num_classes, 10))

batch_size = 20
train_images, train_labels, test_images, test_labels = datasets.mnist()
num_train = train_images.shape[0]
num_complete_batches, leftover = divmod(num_train, batch_size)
num_batches = num_complete_batches + bool(leftover)


def data_stream():
    rng = npr.RandomState(0)
    while True:
        perm = rng.permutation(num_train)
        for i in range(num_batches):
            batch_idx = perm[i * batch_size : (i + 1) * batch_size]
            yield train_images[batch_idx], train_labels[batch_idx]


batches = data_stream()
inputs, outputs = next(batches)


class CNN(nn.Module):
    """Flax CNN Module"""

    @nn.compact
    def __call__(self, x):
        x = nn.Conv(features=32, kernel_size=(3, 3))(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = nn.Conv(features=64, kernel_size=(3, 3))(x)
        x = nn.relu(x)
        x = nn.avg_pool(x, window_shape=(2, 2), strides=(2, 2))
        x = x.reshape((x.shape[0], -1))  # Flatten
        x = nn.Dense(features=256)(x)
        x = nn.relu(x)
        x = nn.Dense(features=10)(x)  # There are 10 classes in MNIST
        x = nn.log_softmax(x)
        return x


class ConvNeuralNetwork(AbstractProblem):
    def __init__(self):
        pass

    @staticmethod
    def objective(params, bparam) -> float:
        def cross_entropy_loss(logits, labels):
            one_hot_labels = jax.nn.one_hot(labels, num_classes=10)
            return -np.mean(np.sum(one_hot_labels * logits, axis=-1))

        logits = CNN().apply({"params": params[0]}, inputs)
        loss = cross_entropy_loss(logits, outputs)
        loss += l2_norm(params) + l2_norm(bparam)
        # vectorization of mini-batch of data.
        # #3rd argument's 0th-axis is vmaped. --> inputs(10)
        # batched_predict = vmap(neural_net_predict, in_axes=(None, None, 0))
        return loss

    @staticmethod
    def init_network_params(key):
        init_shape = np.ones((1, 28, 28, 1), np.float32)
        initial_params = CNN().init(key, init_shape)["params"]
        return initial_params

    def initial_value(self) -> Tuple:
        bparam = [np.array([-0.50], dtype=np.float64)]
        state = [self.init_network_params(random.PRNGKey(0))]
        return state, bparam

    def initial_values(self) -> Tuple:
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.05, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams



from abc import ABC, abstractmethod
import jax.numpy as np
from jax import grad, jit, lax
from jax.experimental.optimizers import l2_norm
from cjax.utils.math_trees import pytree_dot, pytree_sub
from functools import partial


class AbstractProblem(ABC):
    @staticmethod
    @abstractmethod
    def objective(params, bparam, batch_input) -> float:
        pass

    @staticmethod
    @abstractmethod
    def accuracy(params, bparam, batch_input) -> float:
        pass

    @abstractmethod
    def initial_value(self):
        pass

    @abstractmethod
    def initial_values(self):
        pass


class ProblemWraper:
    def __init__(self, problem_object: AbstractProblem):
        self.problem_object = problem_object
        self.objective = self.problem_object.objective
        self.accuracy_fn = self.problem_object.accuracy
        self.initial_value_func = self.problem_object.initial_value
        self.initial_values_func = self.problem_object.initial_values
        self.HPARAMS_PATH = problem_object.HPARAMS_PATH

    def dual_objective(
        self,
        params: list,
        bparam: list,
        lagrange_multiplier: float,
        c2: list,
        secant: list,
        batch_input=0.0,
        delta_s=0.02,
    ) -> float:
        return np.mean(
            self.objective(params, bparam, batch_input)
            + (
                np.multiply(
                    lagrange_multiplier,
                    self.normal_vector(params, bparam, c2, secant, delta_s),
                )
            )
        )

    def initial_value(self):
        return self.initial_value_func()

    def initial_values(self):
        return self.initial_values_func()

    @staticmethod
    def normal_vector(
        params: list, bparams: list, secant_guess: list, secant_vec: list, delta_s
    ) -> float:
        """"""
        result = 0.0
        state_stack = dict()
        state_stack.update({"state": params})
        state_stack.update({"bparam": bparams})
        parc_vec = pytree_sub(state_stack, secant_guess)
        result += pytree_dot(parc_vec, secant_vec)
        #return lax.cond(abs(result) <= 0.1, 0.0, result-delta_s, None)
        return result - delta_s

    def objective_grad(self, params, bparam):  # TODO: JIT?
        grad_J = grad(self.objective, [0, 1])
        params_grad, bparam_grad = grad_J(params, bparam)
        result = l2_norm(params_grad) + l2_norm(bparam_grad)
        return result

    def dual_objective_grad(
        self,
        params: list,
        bparam: list,
        lagrange_multiplier: float,
        c2: list,
        secant: list,
        delta_s=0.02,
    ) -> float:
        return np.mean(
            self.objective_grad(params, bparam)
            + (
                lagrange_multiplier
                * self.normal_vector(params, bparam, c2, secant, delta_s)
            )
        )

    @staticmethod
    def reparm_bijection(params):
        b = 2.0  # greater equal to  0
        a = -0.6
        params = np.power(np.abs(params + b), a) * params
        return params



"""

@article{sinha2020curriculum,
  title={Curriculum By Smoothing},
  author={Sinha, Samarth and Garg, Animesh and Larochelle, Hugo},
  journal={Advances in Neural Information Processing Systems},
  volume={33},
  year={2020}
}
"""
import os
import scipy.io
import numpy as np
import jax.numpy as jnp
import random

import torch
import torch.utils.data as data
from torchvision import transforms, datasets

# import torch.multiprocessing as multiprocessing
# multiprocessing.set_start_method('spawn')
_DATA = "/opt/ml/tmp/jax_example_data/"


class JaxDataWrapper(torch.utils.data.Dataset):
    def __init__(self, train, args, seed=0):
        super().__init__()
        if args["dataset"] == "mnist":
            transform = transforms.Compose(
                [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
            )
            self._data = datasets.MNIST(
                root=args["data"], train=train, download=True, transform=transform
            )
        elif args["dataset"] == "cifar10":
            transform = transforms.Compose(
                [
                    transforms.Scale(32),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (
                            0.5,
                            0.5,
                            0.5,
                        ),
                        (
                            0.5,
                            0.5,
                            0.5,
                        ),
                    ),
                ]
            )
            self._data = datasets.CIFAR10(
                root=args["data"], train=train, download=True, transform=transform
            )
        elif args["dataset"] == "cifar100":
            transform = transforms.Compose(
                [
                    transforms.Scale(32),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (
                            0.5,
                            0.5,
                            0.5,
                        ),
                        (
                            0.5,
                            0.5,
                            0.5,
                        ),
                    ),
                ]
            )
            self._data = datasets.CIFAR100(
                root=args["data"], train=train, download=True, transform=transform
            )
        elif args["dataset"] == "imagenet":
            transform = transforms.Compose(
                [
                    transforms.Scale(64),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (
                            0.5,
                            0.5,
                            0.5,
                        ),
                        (
                            0.5,
                            0.5,
                            0.5,
                        ),
                    ),
                ]
            )
            if train:
                self._data = datasets.ImageFolder(
                    os.path.join(args["data"], "tiny-imagenet-200", "train"),
                    transform=transform,
                )
            else:
                self._data = datasets.ImageFolder(
                    os.path.join(args["data"], "tiny-imagenet-200", "val"),
                    transform=transform,
                )
        else:
            self._data = datasets.MNIST(root=args["data"], train=False, download=True)

        self._data_len = len(self._data)

    def __getitem__(self, index):
        img, label = self._data[index]
        return jnp.asarray(np.asarray(img)), label

    def __len__(self):
        return self._data_len


def jax_collate_fn(batch):
    if isinstance(batch[0], jnp.ndarray):
        return jnp.stack(batch)
    elif isinstance(batch[0], (tuple, list)):
        return type(batch[0])(jax_collate_fn(samples) for samples in zip(*batch))
    else:
        return jnp.asarray(batch)


def get_data(**args):
    args.update({"data": _DATA})
    train_data = JaxDataWrapper(
        train=True,
        args=args,
    )
    test_data = JaxDataWrapper(
        train=False,
        args=args,
    )

    train_loader = data.DataLoader(
        train_data,
        batch_size=args["batch_size"],
        pin_memory=True,
        collate_fn=jax_collate_fn,
        num_workers=args["num_workers"],
        shuffle=True,
        drop_last=True,
    )

    test_loader = data.DataLoader(
        test_data,
        batch_size=args["batch_size"],
        pin_memory=True,
        collate_fn=jax_collate_fn,
        num_workers=args["num_workers"],
        shuffle=True,
        drop_last=False,
    )
    if args["train_only"]:
        return train_loader
    elif args["test_only"]:
        return test_loader
    else:
        return train_loader, test_loader


if __name__ == "__main__":
    train_loader = get_data(
        dataset="mnist",
        batch_size=2,
        data=_DATA,
        num_workers=2,
        train_only=True,
        test_only=False,
    )



"""
author: Paul Bruillard, harsh
"""

import jax.numpy as jnp
from cjax.utils.math_trees import *
from typing import Any


def get_rotation_pytree(src: Any, dst: Any) -> Any:
    """
    Takes two n-dimensional vectors/Pytree and returns an
    nxn rotation matrix mapping cjax to dst.
    Raises Value Error when unsuccessful.
    """

    def __assert_rotation(R):
        if R.ndim != 2:
            print("R must be a matrix")
        a, b = R.shape
        if a != b:
            print("R must be square")
        if (
            not jnp.isclose(jnp.abs(jnp.eye(a) - jnp.dot(R, R.T)).max(), 0.0, rtol=0.5)
        ) or (
            not jnp.isclose(jnp.abs(jnp.eye(a) - jnp.dot(R.T, R)).max(), 0.0, rtol=0.5)
        ):
            print("R is not diagonal")

    if not pytree_shape_array_equal(src, dst):
        print("cjax and dst must be 1-dimensional arrays with the same shape.")

    x = pytree_normalized(src)
    y = pytree_normalized(dst)
    n = len(dst)

    # compute angle between x and y in their spanning space
    theta = jnp.arccos(jnp.dot(x, y))  # they are normalized so there is no denominator
    if jnp.isclose(theta, 0):
        print("x and y are co-linear")
    # construct the 2d rotation matrix connecting x to y in their spanning space
    R = jnp.array([[jnp.cos(theta), -jnp.sin(theta)], [jnp.sin(theta), jnp.cos(theta)]])
    __assert_rotation(R)
    # get projections onto Span<x,y> and its orthogonal complement
    u = x
    v = pytree_normalized(pytree_sub(y, (jnp.dot(u, y) * u)))
    P = jnp.outer(u, u.T) + jnp.outer(
        v, v.T
    )  # projection onto 2d space spanned by x and y
    Q = jnp.eye(n) - P  # projection onto the orthogonal complement of Span<x,y>
    # lift the rotation matrix into the n-dimensional space
    uv = jnp.hstack((u[:, None], v[:, None]))

    R = Q + jnp.dot(uv, jnp.dot(R, uv.T))
    __assert_rotation(R)
    if jnp.any(jnp.logical_not(jnp.isclose(jnp.dot(R, x), y, rtol=0.25))):
        print("Rotation matrix did not work")
    return R


def get_rotation_array(src: Any, dst: Any) -> Any:
    """
    Takes two n-dimensional vectors and returns an
    nxn rotation matrix mapping cjax to dst.
    Raises Value Error when unsuccessful.
    """

    def __assert_rotation(R):
        if R.ndim != 2:
            raise ValueError("R must be a matrix")
        a, b = R.shape
        if a != b:
            raise ValueError("R must be square")
        if (not np.isclose(np.abs(np.eye(a) - np.dot(R, R.T)).max(), 0)) or (
            not np.isclose(np.abs(np.eye(a) - np.dot(R.T, R)).max(), 0)
        ):
            raise ValueError("R is not diagonal")

    def __normalize(x):
        return x / np.sqrt(np.sum(x ** 2))

    if src.shape != dst.shape or src.ndim != 1:
        raise ValueError(
            "cjax and dst must be 1-dimensional arrays with the same shape."
        )
    x = __normalize(src.copy())
    y = __normalize(dst.copy())

    # compute angle between x and y in their spanning space
    theta = np.arccos(np.dot(x, y))  # they are normalized so there is no denominator
    if np.isclose(theta, 0):
        raise ValueError("x and y are co-linear")
    # construct the 2d rotation matrix connecting x to y in their spanning space
    R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    __assert_rotation(R)
    # get projections onto Span<x,y> and its orthogonal complement
    u = x
    v = __normalize((y - (np.dot(u, y) * u)))
    P = np.outer(u, u.T) + np.outer(
        v, v.T
    )  # projection onto 2d space spanned by x and y
    Q = np.eye(n) - P  # projection onto the orthogonal complement of Span<x,y>
    # lift the rotation matrix into the n-dimensional space
    uv = np.hstack((u[:, None], v[:, None]))
    R = Q + np.dot(uv, np.dot(R, uv.T))
    __assert_rotation(R)
    if np.any(np.logical_not(np.isclose(np.dot(R, x), y))):
        raise ValueError("Rotation matrix did not work")
    return R


def projection_affine(n_dim, u, n, u_0):
    """

    Args:
        n_dim: affine transformation space
        u: random point to be projected on n as L
        n: secant normal vector
        u_0: secant starting point

    Returns:

    """
    n_norm = l2_norm(n)
    I = jnp.eye(n_dim)

    p2 = [0 * k for k in range(n_dim)]
    for k in range(n_dim):
        p2[k] = (jnp.dot(n, I[k]) / n_norm ** 2) * n

    p2 = jnp.asarray([p2[i] for i in range(n_dim)])
    u_0 = u_0.reshape(n_dim, 1)
    I = jnp.eye(n_dim)
    t1 = jnp.block([[I, u_0], [jnp.zeros(shape=(1, n_dim)), 1.0]])
    t2 = jnp.block(
        [[p2, jnp.zeros(shape=(n_dim, 1))], [jnp.zeros(shape=(1, n_dim)), 1.0]]
    )
    t3 = jnp.block([[I, -1 * u_0], [jnp.zeros(shape=(1, n_dim)), 1.0]])
    P = jnp.matmul(jnp.matmul(t1, t2), t3)
    pr = jnp.matmul(P, jnp.hstack([u, 1.0]))
    pr = lax.slice(pr, [0], [n_dim])
    return pr


if __name__ == "__main__":
    n = 5
    # key = random.PRNGKey(10)
    # k1, k2 = random.split(key, 2)
    # cjax = random.normal(k1, [n])
    # dst = random.normal(k2, [n])
    # R = get_rotation_pytree(cjax, dst)
    # transformed_vector = np.dot(R, cjax)
    # print(jnp.dot(transformed_vector, dst))

    # n = 3
    # cjax= np.array([0.0,0.0,1.0])
    # dst = np.array([3.0,3.0,3.0])
    # sample = np.array([1.0,2.0,0.0])
    # R = get_rotation_array(cjax, dst)
    # transformed_vector = np.dot(R, sample) + dst
    # print(transformed_vector)



import numpy as onp
from jax.experimental.optimizers import l2_norm
from numpy import random as npr
from cjax.utils.math_trees import pytree_sub

def exp_decay(epoch, initial_lrate):
    k = 0.02
    lrate = initial_lrate * onp.exp(-k * epoch)
    return lrate

def center_data(X):
    mean_x = onp.mean(X, axis=0, keepdims=True)
    reduced_mean = onp.subtract(X, mean_x)
    reduced_mean = reduced_mean.astype(onp.float32)
    return reduced_mean

def grouper(iterable, threshold=0.01):
    prev = None
    group = []
    for item in iterable:
        if not prev or l2_norm(pytree_sub(item, prev)) <= threshold:
            group.append(item)
        else:
            yield group
            group = [item]
        prev = item
    if group:
        yield group


def running_mean(x, N):
    cumsum = onp.cumsum(onp.insert(x, 0, 0))
    return (cumsum[N:] - cumsum[:-N]) / float(N)


def get_cheapest_ant(ants_norm_grads, ants_loss_values, local_test="loss"):
    m = min(ants_norm_grads)
    g_indicies = [i for i, j in enumerate(ants_norm_grads) if j == m]
    print(f"Number of minima by norm: {len(g_indicies)}")

    m = min(ants_loss_values)
    indicies = [i for i, j in enumerate(ants_loss_values) if j == m]
    print(f"Number of minima by loss: {len(indicies)}")
    if local_test == "ma_loss":
        index = npr.randint(0, len(indicies))
        cheapest = indicies[index]
    elif local_test == "norm_gradients":
        index = npr.randint(0, len(g_indicies))
        cheapest = g_indicies[index]
    else:
        index = npr.randint(0, len(indicies))
        cheapest = indicies[index]
    return cheapest



from cjax.continuation.methods.corrector.constrained_corrector import (
    ConstrainedCorrector,
    Corrector,
)
from typing import Tuple
from jax.tree_util import tree_map
from jax import random
from cjax.utils.rotation_ndims import *
from jax import jit
import math
from cjax.optimizer.optimizer import OptimizerCreator
from jax.experimental.optimizers import clip_grads
import numpy.random as npr
from examples.torch_data import get_data
from cjax.utils.datasets import get_mnist_data, meta_mnist  # faster version


class PerturbedFixedCorrecter(Corrector):
    """Minimize the objective using gradient based method along with some constraint and noise"""

    def __init__(
        self,
        objective,
        dual_objective,
        value_fn,
        concat_states,
        key_state,
        compute_min_grad_fn,
        compute_grad_fn,
        hparams,
        pred_state,
        pred_prev_state,
        counter,
    ):
        self.concat_states = concat_states
        self._state = None
        self._bparam = None
        self.opt = OptimizerCreator(
            opt_string=hparams["meta"]["optimizer"], learning_rate=hparams["descent_lr"]
        ).get_optimizer()
        self.objective = objective
        self.dual_objective = dual_objective
        self._lagrange_multiplier = hparams["lagrange_init"]
        self._state_secant_vector = None
        self._state_secant_c2 = None
        self.delta_s = hparams["delta_s"]
        self.descent_period = hparams["descent_period"]
        self.max_norm_state = hparams["max_bounds"]
        self.hparams = hparams
        self.compute_min_grad_fn = compute_min_grad_fn
        self.compute_grad_fn = compute_grad_fn
        self._assign_states()
        self._parc_vec = None
        self.state_stack = dict()
        self.key_state = key_state
        self.pred_state = pred_state
        self.pred_prev_state = pred_prev_state
        self.sphere_radius = hparams["sphere_radius"]
        self.counter = counter
        self.value_fn = value_fn
        # self.data_loader = iter(get_data(dataset=hparams["meta"]['dataset'],
        #                             batch_size=hparams['batch_size'],
        #                             num_workers=hparams['data_workers'],
        #                             train_only=True, test_only=False))
        if hparams["meta"]["dataset"] == "mnist":
            self.data_loader = iter(
                get_mnist_data(
                    batch_size=hparams["batch_size"], resize=hparams["resize_to_small"]
                )
            )
            self.num_batches = meta_mnist(hparams["batch_size"])["num_batches"]
        else:
            self.data_loader = None
            self.num_batches = 1

    def _assign_states(self):
        self._state = self.concat_states[0]
        self._bparam = self.concat_states[1]
        self._state_secant_vector = self.concat_states[2]
        self._state_secant_c2 = self.concat_states[3]

    @staticmethod
    @jit
    def exp_decay(epoch, initial_lrate):
        k = 0.02
        lrate = initial_lrate * np.exp(-k * epoch)
        return lrate

    @staticmethod
    def _perform_perturb_by_projection(
        _state_secant_vector,
        _state_secant_c2,
        key,
        pred_prev_state,
        _state,
        _bparam,
        sphere_radius,
    ):
        ### Secant normal
        n, sample_unravel = pytree_to_vec(
            [_state_secant_vector["state"], _state_secant_vector["bparam"]]
        )
        n = pytree_normalized(n)
        ### sample a random poin in Rn
        # u = tree_map(
        #     lambda a: a + random.uniform(key, a.shape),
        #     pytree_zeros_like(n),
        # )
        print(key)
        u = tree_map(
            lambda a: a + random.normal(key, a.shape),
            pytree_ones_like(n),
        )
        tmp, _ = pytree_to_vec([_state_secant_c2["state"], _state_secant_c2["bparam"]])

        # select a point on the secant normal
        u_0, _ = pytree_to_vec(pred_prev_state)
        # compute projection
        proj_of_u_on_n = projection_affine(len(n), u, n, u_0)

        point_on_plane = u + pytree_sub(tmp, proj_of_u_on_n)  ## state= pred_state + n
        # inv_vec = np.array([-1.0, 1.0])
        parc = pytree_element_mul(
            pytree_normalized(pytree_sub(point_on_plane, tmp)),
            1.0,  # inv_vec[(counter % 2)],
        )
        point_on_plane_2 = tmp + sphere_radius * parc
        print("point on plane ", point_on_plane_2)
        new_sample = sample_unravel(point_on_plane_2)
        state_stack = {}
        state_stack.update({"state": new_sample[0]})
        state_stack.update({"bparam": new_sample[1]})
        _parc_vec = pytree_sub(state_stack, _state_secant_c2)
        return _parc_vec, state_stack

    def _evaluate_perturb(self):
        """Evaluate weather the perturbed vector is orthogonal to secant vector"""

        dot = pytree_dot(
            pytree_normalized(self._parc_vec),
            pytree_normalized(self._state_secant_vector),
        )
        if math.isclose(dot, 0.0, abs_tol=0.25):
            print(f"Perturb was near arc-plane. {dot}")
            self._state = self.state_stack["state"]
            self._bparam = self.state_stack["bparam"]
        else:
            print(f"Perturb was not on arc-plane.{dot}")

    def correction_step(self) -> Tuple:
        """Given the current state optimize to the correct state.

        Returns:
          (state: problem parameters, bparam: continuation parameter) Tuple
        """
        _, key = random.split(random.PRNGKey(self.key_state + npr.randint(1, 100)))
        del _
        quality = 1.0
        N_opt = 10
        stop = False
        corrector_omega = 1.0
        # bparam_grads = pytree_zeros_like(self._bparam)
        print("the radius", self.sphere_radius)
        self._parc_vec, self.state_stack = self._perform_perturb_by_projection(
            self._state_secant_vector,
            self._state_secant_c2,
            key,
            self.pred_prev_state,
            self._state,
            self._bparam,
            self.sphere_radius,
        )
        if self.hparams["_evaluate_perturb"]:
            self._evaluate_perturb()  # does every time

        for j in range(self.descent_period):
            for b_j in range(self.num_batches):
                if self.hparams["meta"]["dataset"] == "mnist":  # TODO: make it generic
                    batch_data = next(self.data_loader)
                else:
                    batch_data = None
                # grads = self.compute_grad_fn(self._state, self._bparam, batch_data)
                # self._state = self.opt.update_params(self._state, grads[0])
                state_grads, bparam_grads = self.compute_min_grad_fn(
                    self._state,
                    self._bparam,
                    self._lagrange_multiplier,
                    self._state_secant_c2,
                    self._state_secant_vector,
                    batch_data,
                    self.delta_s,
                )

                if self.hparams["adaptive"]:
                    self.opt.lr = self.exp_decay(j, self.hparams["natural_lr"])
                    quality = l2_norm(state_grads)  # +l2_norm(bparam_grads)
                    if quality > self.hparams["quality_thresh"]:
                        pass
                        # print(f"quality {quality}, {self.opt.lr}, {bparam_grads} ,{j}")
                    else:
                        if N_opt > (j + 1):  # To get around folds slowly
                            corrector_omega = min(N_opt / (j + 1), 2.0)
                        else:
                            corrector_omega = max(N_opt / (j + 1), 0.5)
                        stop = True
                        print(f"quality {quality} stopping at , {j}th step")
                    state_grads = clip_grads(state_grads, self.hparams["max_clip_grad"])
                    bparam_grads = clip_grads(
                        bparam_grads, self.hparams["max_clip_grad"]
                    )

                self._bparam = self.opt.update_params(self._bparam, bparam_grads, j)
                self._state = self.opt.update_params(self._state, state_grads, j)
                if stop:
                    break
            if stop:
                break

        value = self.value_fn(
            self._state, self._bparam, batch_data
        )  # Todo: why only final batch data
        return self._state, self._bparam, quality, value, corrector_omega



"""
Original code from: https://github.com/google/jax/tree/master/examples
"""
import time

import numpy.random as npr

from jax import jit, grad, hessian
from jax.scipy.special import logsumexp
from jax.flatten_util import ravel_pytree
import jax.numpy as jnp
from cjax.utils.datasets import mnist


def init_random_params(scale, layer_sizes, rng=npr.RandomState(0)):
  return [(scale * rng.randn(m, n), scale * rng.randn(n))
          for m, n, in zip(layer_sizes[:-1], layer_sizes[1:])]

def predict(params, inputs):
  activations = inputs
  for w, b in params[:-1]:
    outputs = jnp.dot(activations, w) + b
    activations = jnp.tanh(outputs)

  final_w, final_b = params[-1]
  logits = jnp.dot(activations, final_w) + final_b
  return logits - logsumexp(logits, axis=1, keepdims=True)

def loss(params, batch):
  inputs, targets = batch
  preds = predict(params, inputs)
  return -jnp.mean(jnp.sum(preds * targets, axis=1))

def accuracy(params, batch):
  inputs, targets = batch
  target_class = jnp.argmax(targets, axis=1)
  predicted_class = jnp.argmax(predict(params, inputs), axis=1)
  return jnp.mean(predicted_class == target_class)


if __name__ == "__main__":
  layer_sizes = [36,10]
  param_scale = 0.1
  step_size = 0.001
  num_epochs = 10000
  batch_size = 5000

  train_images, train_labels, test_images, test_labels = mnist(resize=True,permute_train=False)
  num_train = train_images.shape[0]
  num_complete_batches, leftover = divmod(num_train, batch_size)
  num_batches = num_complete_batches + bool(leftover)

  def data_stream():
    rng = npr.RandomState(0)
    while True:
      perm = rng.permutation(num_train)
      for i in range(num_batches):
        batch_idx = perm[i * batch_size:(i + 1) * batch_size]
        yield train_images[batch_idx], train_labels[batch_idx]
  batches = data_stream()

  @jit
  def update(params, batch):
    grads = grad(loss)(params, batch)
    return [(w - step_size * dw, b - step_size * db)
            for (w, b), (dw, db) in zip(params, grads)]

  params = init_random_params(param_scale, layer_sizes)
  for epoch in range(num_epochs):
    start_time = time.time()
    for _ in range(num_batches):
      params = update(params, next(batches))
    epoch_time = time.time() - start_time

    train_acc = accuracy(params, (train_images, train_labels))
    test_acc = accuracy(params, (test_images, test_labels))
    if epoch%100==0:
        print("Epoch {} in {:0.2f} sec".format(epoch, epoch_time))
        print("Training set accuracy {}".format(train_acc))
        print("Test set accuracy {}".format(test_acc))
  H = hessian(loss)(params, (test_images, test_labels))
  h, _ = ravel_pytree(H)
  eigen_vals = jnp.linalg.eigvals(h.reshape(370, 370)).real
  eigen_vals = sorted(eigen_vals, reverse=True)
  print(eigen_vals)
  neg_count = len(list(filter(lambda x: (x < 0), eigen_vals)))

  # we can also do len(list1) - neg_count
  pos_count = len(list(filter(lambda x: (x >= 0), eigen_vals)))

  print("Positive numbers in the eigen: ", pos_count)
  print("Negative numbers in the eigen: ", neg_count)

  # outputs:
  #Positive numbers in the eigen:  194
  #Negative numbers in the eigen:  176


import jax.numpy as np
from jax.experimental import stax
from jax.nn.initializers import zeros
from jax.experimental.stax import (
    Dense,
    Sigmoid,
    Conv,
    BatchNorm,
    GeneralConv,
    Flatten,
    LogSoftmax,
    Relu,
)
import numpy.random as npr
from jax import random
from cjax.utils.abstract_problem import AbstractProblem
from jax.tree_util import tree_map
from cjax.utils.custom_nn import HomotopyDropout
from cjax.optimizer.optimizer import GDOptimizer, AdamOptimizer
from jax.experimental.optimizers import l2_norm
from cjax.utils.datasets import mnist
from jax import jit, grad
import json
import math
from cjax.utils.datasets import get_mnist_data, meta_mnist
from cjax.utils.math_trees import pytree_element_add
from cjax.utils.evolve_utils import running_mean

data_size = 40000
input_shape = (data_size, 36)
npr.seed(7)
num_classes = 10


def synth_batches():
    while True:
        images = npr.rand(*input_shape).astype("float32")
        yield images


# batches = synth_batches()
# inputs = next(batches)

# train_images, labels, _, _ = mnist(permute_train=True, resize=True)
# del _
# inputs = train_images[:data_size]
#
# del train_images

# u, s, v_t = onp.linalg.svd(inputs, full_matrices=False)
# I = np.eye(v_t.shape[-1])
# I_add = npr.normal(0.0, 0.002, size=I.shape)
# noisy_I = I + I_add


init_fun, conv_net = stax.serial(
    Conv(32, (5, 5), (2, 2), padding="SAME"),
    BatchNorm(),
    Relu,
    Conv(10, (3, 3), (2, 2), padding="SAME"),
    Relu,
    Flatten,
    Dense(num_classes),
    LogSoftmax,
)
_, key = random.split(random.PRNGKey(0))


class DataTopologyAE(AbstractProblem):
    def __init__(self):
        self.HPARAMS_PATH = "hparams.json"

    @staticmethod
    def objective(params, bparam, batch) -> float:
        x, _ = batch
        x = np.reshape(x, (x.shape[0], -1))
        logits = predict_fun(params, x, bparam=bparam[0], rng=key)
        keep = random.bernoulli(key, bparam[0], x.shape)
        inputs_d = np.where(keep, x, 0)

        loss = np.mean(np.square((np.subtract(logits, inputs_d))))
        loss += 1e-8 * (l2_norm(params) + 1 / l2_norm(bparam))
        return loss

    def initial_value(self):
        ae_shape, ae_params = init_fun(random.PRNGKey(0), input_shape)
        assert ae_shape == input_shape
        bparam = [np.array([0.01], dtype=np.float64)]
        return ae_params, bparam

    def initial_values(self):
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.084, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.03, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams


def exp_decay(epoch, initial_lrate):
    k = 0.02
    lrate = initial_lrate * np.exp(-k * epoch)
    return lrate


if __name__ == "__main__":
    problem = DataTopologyAE()
    ae_params, bparam = problem.initial_value()
    bparam = pytree_element_add(bparam, 0.99)
    data_loader = iter(get_mnist_data(batch_size=25000, resize=True))
    num_batches = meta_mnist(batch_size=25000)["num_batches"]
    print(f"num of bathces: {num_batches}")
    compute_grad_fn = jit(grad(problem.objective, [0]))

    with open(problem.HPARAMS_PATH, "r") as hfile:
        hparams = json.load(hfile)
    opt = AdamOptimizer(learning_rate=hparams["descent_lr"])
    ma_loss = []
    for epoch in range(500):
        for b_j in range(num_batches):
            batch = next(data_loader)
            grads = compute_grad_fn(ae_params, bparam, batch)
            ae_params = opt.update_params(ae_params, grads[0], step_index=epoch)
            loss = problem.objective(ae_params, bparam, batch)
            ma_loss.append(loss)
            print(f"loss:{loss}  norm:{l2_norm(grads)}")
        opt.lr = exp_decay(epoch, hparams["descent_lr"])
        if len(ma_loss) > 40:
            loss_check = running_mean(ma_loss, 30)
            if math.isclose(
                loss_check[-1], loss_check[-2], abs_tol=hparams["loss_tol"]
            ):
                print(f"stopping at {epoch}")
                break

    train_images, train_labels, test_images, test_labels = mnist(
        permute_train=False, resize=True
    )
    val_loss = problem.objective(ae_params, bparam, (test_images, test_labels))
    print(f"val loss: {val_loss}")

    # init_c = constant_2d(I)
    # print(init_c(key=0, shape=(8,8)))



from cjax.utils.datasets import mnist_raw, center_data, img_resize, _partial_flatten, _one_hot
import numpy as np
import cv2
import numpy.random as npr

def mnist_gamma(resize=True, filter=False, center=True):

    train_images, train_labels, test_images, test_labels = mnist_raw()
    if filter:
        train_filter = np.where(train_labels == 1)
        train_images = train_images[train_filter]
        train_labels = train_labels[train_filter]

        test_filter = np.where(test_labels == 1)
        test_images = test_images[test_filter]
        test_labels = test_labels[test_filter]

    if resize:
        train_images = img_resize(train_images)
        test_images = img_resize(test_images)

    train_dict = {}
    alters = np.linspace(start=0.05, stop=0.95, num=9)
    for k in range(9):
        train_images_tmp = []
        for img in train_images:
            dst = adjust_gamma(img, alters[k])
            train_images_tmp.append(dst)
        train_images_tmp = np.asarray(train_images_tmp)
        train_images_tmp = _partial_flatten(train_images_tmp) / np.float32(255.0)
        if center:
            train_images_tmp = center_data(train_images_tmp)
        train_dict.update({k: train_images_tmp})
    train_images = _partial_flatten(train_images) / np.float32(255.0)
    if center:
        train_images = center_data(train_images)
        test_images = center_data(test_images)
    train_dict.update({9: train_images})

    test_images = _partial_flatten(test_images) / np.float32(255.0)
    train_labels = _one_hot(train_labels, 10)
    test_labels = _one_hot(test_labels, 10)
    del train_images
    return train_dict, train_labels, test_images, test_labels

def adjust_gamma(image, gamma=1.0):
    if (gamma<=0.05):
        gamma = 0.05
    invGamma = 1.0/gamma
    table = np.array([
      ((i / 255.0) ** invGamma) * 255
      for i in np.arange(0, 256)])
    return cv2.LUT(image.astype(np.uint8), table.astype(np.uint8))


def get_mnist_batch_alter(train_images_dict: dict, train_labels, test_images,
                    test_labels, alter: list, batch_size, resize, filter=False):
    # alter gamma in image
    alter = alter[0]
    k = int(alter*10)
    if k>9:
        k=9
    train_images = train_images_dict[k]
    print("batch_size: ",batch_size, train_images[0].shape)
    total_data_len = train_images.shape[0]
    num_complete_batches, leftover = divmod(total_data_len, batch_size)
    num_batches = num_complete_batches + bool(leftover)
    rng = npr.RandomState(0)
    while True:
        perm = rng.permutation(total_data_len)
        for i in range(num_batches):
            batch_idx = perm[i * batch_size : (i + 1) * batch_size]
            yield train_images[batch_idx], train_labels[batch_idx]





from cjax.continuation.base_continuation import Continuation
from cjax.continuation.states.state_variables import StateVariable, StateWriter
from cjax.optimizer.optimizer import OptimizerCreator
from cjax.continuation.methods.predictor.arc_secant_predictor import SecantPredictor
from cjax.utils.data_img_gamma import mnist_gamma
from cjax.utils.datasets import mnist, get_mnist_data, meta_mnist
from cjax.continuation.methods.corrector.unconstrained_corrector import (
    UnconstrainedCorrector,
)
from jax.tree_util import *
import gc
from cjax.utils.profiler import profile
from jax import jit, grad
from cjax.utils.math_trees import *
import jax.numpy as np
import mlflow


class SecantContinuation(Continuation):
    """Secant Continuation strategy.

    Composed of natural predictor and unconstrained corrector"""

    def __init__(self, state, bparam, state_0, bparam_0, counter, objective, accuracy_fn, hparams):
        self._state_wrap = StateVariable(state, counter)
        self._bparam_wrap = StateVariable(bparam, counter)
        self._prev_state = state_0
        self._prev_bparam = bparam_0
        self.objective = objective
        self.accuracy_fn = accuracy_fn
        self.value_func = jit(self.objective)
        self._value_wrap = StateVariable(0.005, counter)
        self._quality_wrap = StateVariable(0.005, counter)
        self.sw = None
        self.hparams = hparams
        self.opt = OptimizerCreator(
            opt_string=hparams["meta"]["optimizer"], learning_rate=hparams["natural_lr"]
        ).get_optimizer()
        if hparams["meta"]["dataset"] == "mnist":
            if hparams["continuation_config"] == 'data':
                self.dataset_tuple = mnist_gamma(
                    resize=hparams["resize_to_small"],
                    filter=hparams["filter"])
            else:
                self.dataset_tuple = mnist(
                    resize=hparams["resize_to_small"],
                    filter=hparams["filter"])
        self.continuation_steps = hparams["continuation_steps"]

        self.output_file = hparams["meta"]["output_dir"]
        self._delta_s = hparams["delta_s"]
        self._prev_delta_s = hparams["delta_s"]
        self._omega = hparams["omega"]
        self.grad_fn = jit(grad(self.objective, argnums=[0]))
        self.prev_secant_direction = None

    @profile(sort_by="cumulative", lines_to_print=10, strip_dirs=True)
    def run(self):
        """Runs the continuation strategy.

        A continuation strategy that defines how predictor and corrector components of the algorithm
        interact with the states of the mathematical system.
        """
        self.sw = StateWriter(f"{self.output_file}/version.json")

        for i in range(self.continuation_steps):
            if i == 0 and self.hparams["natural_start"]:
                print(f" unconstrained solver for 1st step")
                concat_states = [
                    self._prev_state,
                    pytree_element_add(self._prev_bparam, 0.05),
                ]

                corrector = UnconstrainedCorrector(
                    objective=self.objective,
                    concat_states=concat_states,
                    grad_fn=self.grad_fn,
                    value_fn=self.value_func,
                    accuracy_fn=self.accuracy_fn,
                    hparams=self.hparams,
                    dataset_tuple=self.dataset_tuple
                )
                state, bparam, quality, value, val_loss = corrector.correction_step()
                self._state_wrap.state = state
                self._bparam_wrap.state = bparam
                del corrector, state, bparam, quality, value, concat_states

            print(self._value_wrap.get_record(), self._bparam_wrap.get_record())
            self._state_wrap.counter = i
            self._bparam_wrap.counter = i
            self._value_wrap.counter = i
            self._quality_wrap.counter = i
            self.sw.write(
                [
                    self._state_wrap.get_record(),
                    self._bparam_wrap.get_record(),
                    self._value_wrap.get_record(),
                    self._quality_wrap.get_record(),
                ]
            )

            concat_states = [
                (self._prev_state, self._prev_bparam),
                (self._state_wrap.state, self._bparam_wrap.state),
                self.prev_secant_direction,
            ]

            predictor = SecantPredictor(
                concat_states=concat_states,
                delta_s=self._delta_s,
                prev_delta_s=self._prev_delta_s,
                omega=self._omega,
                net_spacing_param=self.hparams["net_spacing_param"],
                net_spacing_bparam=self.hparams["net_spacing_bparam"],
                hparams=self.hparams,
            )
            predictor.prediction_step()

            self.prev_secant_direction = predictor.secant_direction

            concat_states = [predictor.state, predictor.bparam]
            del predictor
            gc.collect()

            corrector = UnconstrainedCorrector(
                objective=self.objective,
                concat_states=concat_states,
                grad_fn=self.grad_fn,
                value_fn=self.value_func,
                accuracy_fn=self.accuracy_fn,
                hparams=self.hparams,
                dataset_tuple=self.dataset_tuple
            )
            state, bparam, quality, value, val_loss, val_acc = corrector.correction_step()

            corrector_omega = 0.005 # why fixed check TODO
            self._prev_delta_s = self._delta_s
            self._delta_s = corrector_omega * self._delta_s
            self._delta_s = min(self._delta_s, self.hparams["max_arc_len"])
            self._delta_s = max(self._delta_s, self.hparams["min_arc_len"])

            self._state_wrap.state = state
            self._bparam_wrap.state = bparam
            self._value_wrap.state = value
            self._quality_wrap.state = quality
            del corrector
            del concat_states
            gc.collect()
            if self._bparam_wrap.state[0] >= self.hparams["lambda_max"]:
                self.sw.write(
                    [
                        self._state_wrap.get_record(),
                        self._bparam_wrap.get_record(),
                        self._value_wrap.get_record(),
                        self._quality_wrap.get_record(),
                    ]
                )
                break
            mlflow.log_metrics({
                "train_loss": float(self._value_wrap.state),
                "delta_s": float(self._delta_s),
                "norm grads": float(self._quality_wrap.state),
                "val_loss": float(val_loss)
            }, i)




from cjax.continuation.methods.corrector.constrained_corrector import (
    ConstrainedCorrector,
    Corrector,
)
from typing import Tuple
from jax.tree_util import tree_map
from jax import random
from cjax.utils.rotation_ndims import *
from jax import jit
import math
from cjax.optimizer.optimizer import OptimizerCreator
from jax.experimental.optimizers import clip_grads
import numpy.random as npr
from cjax.utils.evolve_utils import *
import numpy as onp
from examples.torch_data import get_data
from cjax.utils.datasets import get_preload_mnist_data, meta_mnist
from cjax.utils.data_img_gamma import get_mnist_batch_alter


class PerturbedFixedCorrecter(Corrector):
    """Minimize the objective using gradient based method along with some constraint and noise"""

    def __init__(
        self,
        objective,
        dual_objective,
        accuracy_fn1,
        value_fn,
        concat_states,
        key_state,
        compute_min_grad_fn,
        compute_grad_fn,
        hparams,
        delta_s,
        pred_state,
        pred_prev_state,
        counter,
        dataset_tuple
    ):
        self.concat_states = concat_states
        self._state = None
        self._bparam = None
        self.opt = OptimizerCreator(
            opt_string=hparams["meta"]["optimizer"], learning_rate=hparams["descent_lr"]
        ).get_optimizer()
        self.objective = objective
        self.dual_objective = dual_objective
        self._lagrange_multiplier = hparams["lagrange_init"]
        self._state_secant_vector = None
        self._state_secant_c2 = None
        self.delta_s = delta_s
        self.descent_period = hparams["descent_period"]
        self.max_norm_state = hparams["max_bounds"]
        self.hparams = hparams
        self.compute_min_grad_fn = compute_min_grad_fn
        self.compute_grad_fn = compute_grad_fn
        self._assign_states()
        self._parc_vec = None
        self.state_stack = dict()
        self.key_state = key_state
        self.pred_state = pred_state
        self.pred_prev_state = pred_prev_state
        self.sphere_radius = hparams["sphere_radius"]
        self.counter = counter
        self.value_fn = value_fn
        self.accuracy_fn1 = accuracy_fn1
        self.dataset_tuple = dataset_tuple
        if hparams["meta"]["dataset"] == "mnist":
            (self.train_images, self.train_labels,
             self.test_images, self.test_labels) = dataset_tuple

            if hparams["continuation_config"] == 'data':
                # data continuation
                self.data_loader = iter(
                    get_mnist_batch_alter(
                        self.train_images,
                        self.train_labels,
                        self.test_images,
                        self.test_labels,
                        alter=self._bparam,
                        batch_size=hparams["batch_size"],
                        resize=hparams["resize_to_small"],
                        filter=hparams["filter"]
                    )
                )
            else:
                # model continuation
                self.data_loader = iter(
                    get_preload_mnist_data(self.train_images,
                                           self.train_labels,
                                           self.test_images,
                                           self.test_labels,
                                           batch_size=hparams["batch_size"],
                                           resize=hparams["resize_to_small"],
                                           filter=hparams["filter"])
                )
            self.num_batches = meta_mnist(hparams["batch_size"], hparams["filter"])["num_batches"]
        else:
            self.data_loader = None
            self.num_batches = 1

    def _assign_states(self):
        self._state = self.concat_states[0]
        self._bparam = self.concat_states[1]
        self._state_secant_vector = self.concat_states[2]
        self._state_secant_c2 = self.concat_states[3]

    @staticmethod
    @jit
    def exp_decay(epoch, initial_lrate):
        k = 0.02
        lrate = initial_lrate * np.exp(-k * epoch)
        return lrate

    @staticmethod
    @jit
    def _perform_perturb_by_projection(
        _state_secant_vector,
        _state_secant_c2,
        key,
        pred_prev_state,
        _state,
        _bparam,
        counter,
        sphere_radius,
    ):
        ### Secant normal
        n, sample_unravel = pytree_to_vec(
            [_state_secant_vector["state"], _state_secant_vector["bparam"]]
        )
        n = pytree_normalized(n)
        ### sample a random poin in Rn
        # u = tree_map(
        #     lambda a: a + random.uniform(key, a.shape),
        #     pytree_zeros_like(n),
        # )
        u = tree_map(
            lambda a: a + random.normal(key, a.shape),
            pytree_ones_like(n),
        )
        tmp, _ = pytree_to_vec([_state_secant_c2["state"], _state_secant_c2["bparam"]])

        # select a point on the secant normal
        u_0, _ = pytree_to_vec(pred_prev_state)
        # compute projection
        proj_of_u_on_n = projection_affine(len(n), u, n, u_0)

        point_on_plane = u + pytree_sub(tmp, proj_of_u_on_n)  ## state= pred_state + n
        #noise = random.uniform(key, [1], minval=-0.003, maxval=0.03)
        inv_vec = np.array([-1.0, 1.0])
        parc = pytree_element_mul(
            pytree_normalized(pytree_sub(point_on_plane, tmp)),
            inv_vec[(counter % 2)],
        )
        point_on_plane_2 = tmp + sphere_radius * parc
        new_sample = sample_unravel(point_on_plane_2)
        state_stack = {}
        state_stack.update({"state": new_sample[0]})
        state_stack.update({"bparam": new_sample[1]})
        _parc_vec = pytree_sub(state_stack, _state_secant_c2)
        return _parc_vec, state_stack

    def _evaluate_perturb(self):
        """Evaluate weather the perturbed vector is orthogonal to secant vector"""

        dot = pytree_dot(
            pytree_normalized(self._parc_vec),
            pytree_normalized(self._state_secant_vector),
        )
        if math.isclose(dot, 0.0, abs_tol=0.15):
            print(f"Perturb was near arc-plane. {dot}")
        else:
            print(f"Perturb was not on arc-plane.{dot}")

    def correction_step(self) -> Tuple:
        """Given the current state optimize to the correct state.

        Returns:
          (state: problem parameters, bparam: continuation parameter) Tuple
        """

        quality = 1.0

        ants_norm_grads = [5.0 for _ in range(self.hparams["n_wall_ants"])]
        ants_loss_values = [5.0 for _ in range(self.hparams["n_wall_ants"])]
        ants_state = [self._state for _ in range(self.hparams["n_wall_ants"])]
        ants_bparam = [self._bparam for _ in range(self.hparams["n_wall_ants"])]
        for i_n in range(self.hparams["n_wall_ants"]):
            corrector_omega = 1.0
            stop = False
            _, key = random.split(
                random.PRNGKey(self.key_state + i_n + npr.randint(1, (i_n + 1) * 10))
            )
            del _
            self._parc_vec, self.state_stack = self._perform_perturb_by_projection(
                self._state_secant_vector,
                self._state_secant_c2,
                key,
                self.pred_prev_state,
                self._state,
                self._bparam,
                i_n,
                self.sphere_radius,
            )
            if self.hparams["_evaluate_perturb"]:
                self._evaluate_perturb()  # does every time

            ants_state[i_n] = self.state_stack["state"]
            ants_bparam[i_n] = self.state_stack["bparam"]

            if self.hparams["continuation_config"]=="data":
                if self.hparams["meta"]["dataset"] == "mnist":  # TODO: make it generic
                    self.data_loader = iter(get_mnist_batch_alter(
                        self.train_images,
                        self.train_labels,
                        self.test_images,
                        self.test_labels,
                        alter = self._bparam,
                        batch_size=self.hparams["batch_size"],
                        resize=self.hparams["resize_to_small"],
                        filter=self.hparams["filter"]
                    ))
                    batch_data = next(self.data_loader)
                else:
                    batch_data = None
            else:
                if self.hparams["meta"]["dataset"] == "mnist":
                    batch_data = next(self.data_loader)
                else:
                    batch_data = None

            D_values = []
            print(f"num_batches", self.num_batches)
            for j_epoch in range(self.descent_period):
                for b_j in range(self.num_batches):

                    #alternate
                    # grads = self.compute_grad_fn(self._state, self._bparam, batch_data)
                    # self._state = self.opt.update_params(self._state, grads[0])
                    state_grads, bparam_grads = self.compute_min_grad_fn(
                        ants_state[i_n],
                        ants_bparam[i_n],
                        self._lagrange_multiplier,
                        self._state_secant_c2,
                        self._state_secant_vector,
                        batch_data,
                        self.delta_s,
                    )

                    if self.hparams["adaptive"]:
                        self.opt.lr = self.exp_decay(
                            j_epoch, self.hparams["natural_lr"]
                        )
                        quality = l2_norm(state_grads) #l2_norm(bparam_grads)
                        if self.hparams["local_test_measure"] == "norm_gradients":
                            if quality > self.hparams["quality_thresh"]:
                                pass
                                print(
                                    f"quality {quality}, {self.opt.lr}, {bparam_grads} ,{j_epoch}"
                                )
                            else:
                                stop = True
                                print(
                                    f"quality {quality} stopping at , {j_epoch}th step"
                                )
                        else:
                            print(
                                f"quality {quality}, {bparam_grads} ,{j_epoch}"
                            )
                            if len(D_values) >= 20:
                                tmp_means = running_mean(D_values, 10)
                                if (math.isclose(
                                tmp_means[-1],
                                tmp_means[-2],
                                abs_tol=self.hparams["loss_tol"]
                                )):
                                    print(
                                        f"stopping at , {j_epoch}th step, {ants_bparam[i_n]} bparam"
                                    )
                                    stop = True

                        state_grads = clip_grads(
                            state_grads, self.hparams["max_clip_grad"]
                        )
                        bparam_grads = clip_grads(
                            bparam_grads, self.hparams["max_clip_grad"]
                        )

                    if self.hparams["guess_ant_steps"] >= (
                        j_epoch + 1
                    ):  # To get around folds slowly
                        corrector_omega = min(
                            self.hparams["guess_ant_steps"] / (j_epoch + 1), 1.5
                        )
                    else:
                        corrector_omega = max(
                            self.hparams["guess_ant_steps"] / (j_epoch + 1), 0.05
                        )

                    ants_state[i_n] = self.opt.update_params(
                        ants_state[i_n], state_grads, j_epoch
                    )
                    ants_bparam[i_n] = self.opt.update_params(
                        ants_bparam[i_n], bparam_grads, j_epoch
                    )
                    ants_loss_values[i_n] = self.value_fn(
                        ants_state[i_n], ants_bparam[i_n], batch_data
                    )
                    D_values.append(ants_loss_values[i_n])
                    ants_norm_grads[i_n] = quality
                    # if stop:
                    #     break
                    if (
                        self.hparams["meta"]["dataset"] == "mnist"
                    ):  # TODO: make it generic
                        batch_data = next(self.data_loader)
                if stop:
                    break

        # ants_group = dict(enumerate(grouper(ants_state, tolerence), 1))
        # print(f"Number of groups: {len(ants_group)}")
        cheapest_index = get_cheapest_ant(ants_norm_grads, ants_loss_values,
                                          local_test=self.hparams["local_test_measure"])
        self._state = ants_state[cheapest_index]
        self._bparam = ants_bparam[cheapest_index]
        value = self.value_fn(
            self._state, self._bparam, batch_data
        )  # Todo: why only final batch data


        val_loss = self.value_fn(self._state, self._bparam, (self.test_images,self.test_labels))
        print(f"val loss: {val_loss}")
        val_acc = self.accuracy_fn1(self._state, self._bparam, (self.test_images, self.test_labels))
        print(f"val acc: {val_acc}")

        return self._state, self._bparam, quality, value, val_loss, val_acc, corrector_omega



from unittest import TestCase

from cjax.utils.math_trees import *
import jax.numpy as np
from examples.conv_nn.conv_nn import ConvNeuralNetwork
import copy


class TestPytree(TestCase):
    def setUp(self):
        self.x = [np.array([0.5, 0.5]), 0.5, {"beta": 0.5}]
        self.y = [np.array([0.5, 0.5]), 0.5, {"beta": 0.5}]
        self.z = [np.array([0.0, 0.0]), 0.0, {"beta": 0.0}]
        self.x1 = [np.array([1.0, 1.0]), 1.0, {"beta": 1.0}]
        self.cnn_problem = ConvNeuralNetwork()
        self.flax_conv, self.bparams = self.cnn_problem.initial_value()
        self.flax_conv1 = copy.deepcopy(self.flax_conv)


class TestPytreeDot(TestPytree):
    def test_dot(self):
        self.assertEqual(1.0, pytree_dot(self.x, self.y))
        self.assertEqual(pytree_array_equal(self.flax_conv, self.flax_conv1), True)
        self.assertEqual(361.93063, pytree_dot(self.flax_conv, self.flax_conv1))


class TestPytreeOnesLike(TestPytree):
    def test_dot(self):
        self.assertEqual(pytree_array_equal(self.x1, pytree_ones_like(self.x)), True)


class TestPytreeSub(TestPytree):
    def test_sub(self):
        self.assertEqual(pytree_array_equal(self.z, pytree_sub(self.x, self.y)), True)
        self.assertEqual(pytree_array_equal(self.flax_conv, self.flax_conv1), True)
        self.assertEqual(
            pytree_array_equal(
                pytree_zeros_like(self.flax_conv),
                pytree_sub(self.flax_conv, self.flax_conv1),
            ),
            True,
        )



from typing import Tuple

from jax import grad, jit
from jax.experimental.optimizers import l2_norm
from cjax.continuation.methods.corrector.base_corrector import Corrector
from cjax.optimizer.optimizer import OptimizerCreator
from cjax.utils.datasets import meta_mnist, get_preload_mnist_data, mnist
from cjax.utils.data_img_gamma import get_mnist_batch_alter, mnist_gamma
from cjax.utils.evolve_utils import running_mean, exp_decay
import math


class UnconstrainedCorrector(Corrector):
    """Minimize the objective using gradient based method."""

    def __init__(self, objective, concat_states, grad_fn, value_fn, accuracy_fn , hparams, dataset_tuple):
        self.concat_states = concat_states
        self._state = None
        self._bparam = None
        self.opt = OptimizerCreator(
            opt_string=hparams["meta"]["optimizer"], learning_rate=hparams["natural_lr"]
        ).get_optimizer()
        self.objective = objective
        self.accuracy_fn = accuracy_fn
        self.warmup_period = hparams["warmup_period"]
        self.hparams = hparams
        self.grad_fn = grad_fn
        self.value_fn = value_fn
        self._assign_states()
        if hparams["meta"]["dataset"] == "mnist":
            (self.train_images, self.train_labels,
             self.test_images, self.test_labels) = dataset_tuple
            if hparams["continuation_config"] == 'data':
                # data continuation
                self.data_loader = iter(
                    get_mnist_batch_alter(
                        self.train_images,
                        self.train_labels,
                        self.test_images,
                        self.test_labels,
                        alter=self._bparam,
                        batch_size=hparams["batch_size"],
                        resize=hparams["resize_to_small"],
                        filter=hparams["filter"]
                    )
                )
            else:
                # model continuation
                self.data_loader = iter(
                    get_preload_mnist_data(self.train_images,
                                           self.train_labels,
                                           self.test_images,
                                           self.test_labels,
                                           batch_size=hparams["batch_size"],
                                           resize=hparams["resize_to_small"],
                                           filter=hparams["filter"])
                )

            self.num_batches = meta_mnist(hparams["batch_size"], hparams["filter"])["num_batches"]
        else:
            self.data_loader = None
            self.num_batches = 1

    def _assign_states(self):
        self._state, self._bparam = self.concat_states

    def correction_step(self) -> Tuple:
        """Given the current state optimize to the correct state.

        Returns:
          (state: problem parameters, bparam: continuation parameter) Tuple
        """

        quality = 1.0
        ma_loss = []
        stop = False
        print("learn_rate", self.opt.lr)
        for k in range(self.warmup_period):
            for b_j in range(self.num_batches):
                batch = next(self.data_loader)
                grads = self.grad_fn(self._state, self._bparam, batch)
                self._state = self.opt.update_params(self._state, grads[0])
                quality = l2_norm(grads)
                value = self.value_fn(self._state, self._bparam, batch)
                ma_loss.append(value)
                self.opt.lr = exp_decay(k, self.hparams["natural_lr"])
                if self.hparams["local_test_measure"] == "norm_gradients":
                    if quality > self.hparams["quality_thresh"]:
                        pass
                        print(
                            f"quality {quality}, {self.opt.lr} ,{k}"
                        )
                    else:
                        stop = True
                        print(
                            f"quality {quality} stopping at , {k}th step"
                        )
                else:
                    if len(ma_loss) >= 20:
                        tmp_means = running_mean(ma_loss, 10)
                        if math.isclose(
                                tmp_means[-1],
                                tmp_means[-2],
                                abs_tol=self.hparams["loss_tol"],
                        ):
                            print(
                                f"stopping at , {k}th step"
                            )
                            stop = True
            if stop:
                print("breaking")
                break

        val_loss = self.value_fn(self._state, self._bparam, (self.test_images, self.test_labels))
        val_acc = self.accuracy_fn(self._state, self._bparam, (self.test_images, self.test_labels))
        return self._state, self._bparam, quality, value, val_loss, val_acc



from cjax.continuation.methods.predictor.base_predictor import Predictor
from jax.tree_util import tree_multimap
from cjax.utils.math_trees import *
from jax import jit


class SecantPredictor(Predictor):
    def __init__(
        self,
        concat_states,
        delta_s,
        prev_delta_s,
        omega,
        net_spacing_param,
        net_spacing_bparam,
        hparams,
    ):
        super().__init__(concat_states)
        self._prev_state = None
        self._prev_bparam = None
        self.prev_delta_s = prev_delta_s
        self.delta_s = delta_s
        self.omega = omega
        self.net_spacing_param = net_spacing_param
        self.net_spacing_bparam = net_spacing_bparam
        self.secantvar_state = None
        self.secantvar_bparam = None
        self.secant_direction = dict()
        self.prev_secant_direction = None # only for choose direction logic.
        self.hparams = hparams

    def _assign_states(self):
        self._prev_state, self._prev_bparam = self._concat_states[0]
        self._state, self._bparam = self._concat_states[1]
        self.prev_secant_direction = self._concat_states[2]

    @staticmethod
    @jit
    def _compute_secant(
        _state,
        _bparam,
        _prev_state,
        _prev_bparam,
        delta_s,
        prev_delta_s,
    ):
        secant_direction = {}
        state_sub = pytree_sub(_state, _prev_state)
        bparam_sub = pytree_sub(_bparam, _prev_bparam)
        secant_direction.update(
            {"state": pytree_element_mul(state_sub, delta_s / prev_delta_s)}
        )

        secant_direction.update(
            {"bparam": pytree_element_mul(bparam_sub, delta_s / prev_delta_s)}
        )

        return secant_direction

    def prediction_step(self):
        """Given current state predict next state.
        Updates (state_guess: problem parameters, bparam_guess: continuation parameter)
        """
        self._assign_states()

        self.secant_direction = self._compute_secant(
            self._state,
            self._bparam,
            self._prev_state,
            self._prev_bparam,
            self.delta_s,
            self.prev_delta_s,
        )
        # self._choose_direction()
        self._state = tree_multimap(
            lambda a, b: a + b,
            self._state,
            self.secant_direction["state"],
        )
        self._bparam = tree_multimap(
            lambda a, b: a + b,
            self._bparam,
            self.secant_direction["bparam"],
        )

    def get_secant_vector_concat(self):
        """Concatenated secant vector.

        Returns:
          [state_vector: problem parameters, bparam_vector: continuation parameter] list
        """
        return self.secant_direction

    # TODO: below function needed?
    def get_secant_concat(self):
        """Concatenated secant guess/point.

        Returns:
          [state_guess: problem parameters, bparam_guess: continuation parameter] list
        """
        concat = dict()
        concat.update({"state": self._state})
        concat.update({"bparam": self._bparam})
        return concat



"""
Main file to run contination on the user defined problem. Examples can be found in the examples/ directory.


Continuation is topological procedure to train a neural network. This module tracks all
the critical points or fixed points and dumps them to  output file provided in hparams.json file.

  Typical usage example:

  continuation = ContinuationCreator(
            problem=problem, hparams=hparams
        ).get_continuation_method()
        continuation.run()


"""
from cjax.continuation.creator.continuation_creator import ContinuationCreator
from examples.data_cont_ae.autoencoder import DataTopologyAE
from cjax.utils.abstract_problem import ProblemWraper
import json
from jax.config import config
from datetime import datetime
from cjax.utils.visualizer import bif_plot
import torch.multiprocessing as multiprocessing

config.update("jax_debug_nans", True)

# TODO: use **kwargs to reduce params

if __name__ == "__main__":

    multiprocessing.set_start_method("spawn")
    problem = DataTopologyAE()
    problem = ProblemWraper(problem)

    with open(problem.HPARAMS_PATH, "r") as hfile:
        hparams = json.load(hfile)
    start_time = datetime.now()

    if hparams["n_perturbs"] > 1:
        for perturb in range(hparams["n_perturbs"]):
            print(f"Running perturb {perturb}")
            continuation = ContinuationCreator(
                problem=problem, hparams=hparams, key=perturb
            ).get_continuation_method()
            continuation.run()
    else:
        continuation = ContinuationCreator(
            problem=problem, hparams=hparams
        ).get_continuation_method()
        continuation.run()

    end_time = datetime.now()
    print(f"Duration: {end_time-start_time}")



"""
Main file to run contination on the user defined problem. Examples can be found in the examples/ directory.


Continuation is topological procedure to train a neural network. This module tracks all
the critical points or fixed points and dumps them to  output file provided in hparams.json file.

  Typical usage example:

  continuation = ContinuationCreator(
            problem=problem, hparams=hparams
        ).get_continuation_method()
        continuation.run()

"""
from cjax.continuation.creator.continuation_creator import ContinuationCreator
from examples.data_simple_classifier.data_classifier import DataContClassifier
from cjax.utils.abstract_problem import ProblemWraper
import json
from jax.config import config
from datetime import datetime
import mlflow
from cjax.utils.visualizer import pick_array, bif_plot

config.update("jax_debug_nans", True)

# TODO: use **kwargs to reduce params

if __name__ == "__main__":
    problem = DataContClassifier()
    print(f"problem name: {problem.__class__}")
    problem = ProblemWraper(problem)


    with open(problem.HPARAMS_PATH, "r") as hfile:
        hparams = json.load(hfile)
    mlflow.set_tracking_uri(hparams['meta']["mlflow_uri"])
    mlflow.set_experiment(hparams['meta']["name"])

    with mlflow.start_run(run_name=hparams['meta']["method"]+"-"+hparams["meta"]["optimizer"]) as run:
        mlflow.log_dict(hparams, artifact_file="hparams/hparams.json")
        mlflow.log_text("", artifact_file="output/_touch.txt")
        artifact_uri = mlflow.get_artifact_uri("output/")
        hparams["meta"]["output_dir"] = artifact_uri
        print(f"URI: {artifact_uri}")
        start_time = datetime.now()

        if hparams["n_perturbs"] > 1:
            for perturb in range(hparams["n_perturbs"]):
                print(f"Running perturb {perturb}")
                continuation = ContinuationCreator(
                    problem=problem, hparams=hparams, key=perturb
                ).get_continuation_method()
                continuation.run()
        else:
            continuation = ContinuationCreator(
                problem=problem, hparams=hparams
            ).get_continuation_method()
            continuation.run()

        end_time = datetime.now()
        print(f"Duration: {end_time-start_time}")

        figure = bif_plot(hparams["meta"]["output_dir"], pick_array)
        mlflow.log_figure(figure, artifact_file="plots/fig.png")



import jax.numpy as np
import numpy as onp
from jax.experimental import stax
from jax.nn.initializers import zeros, orthogonal, delta_orthogonal, uniform, normal
from jax.nn import sigmoid, hard_tanh, soft_sign, relu, logsumexp
from jax.experimental.stax import Dense, Sigmoid, Softplus, Tanh, LogSoftmax,Relu
from cjax.optimizer.optimizer import GDOptimizer, AdamOptimizer, OptimizerCreator
import numpy.random as npr
from jax import random
from jax import jit, vmap, grad, hessian
from cjax.utils.abstract_problem import AbstractProblem
from jax.tree_util import tree_map
from jax.flatten_util import ravel_pytree
from cjax.utils.math_trees import *
from cjax.utils.evolve_utils import *
from cjax.continuation.states.state_variables import StateWriter
import math
import mlflow
import json
from cjax.utils.custom_nn import constant_2d, HomotopyDense, v_2d
from cjax.utils.datasets import mnist, get_mnist_data, meta_mnist
import pickle
from examples.torch_data import get_data

npr.seed(7)
orth_init_cont = False
input_shape = (30000, 36)
def accuracy(params, bparams, batch):
    inputs, targets = batch
    target_class = np.argmax(targets, axis=-1)
    predicted_class = np.argmax(predict_fun(params, inputs), axis=-1)
    return np.mean(predicted_class == target_class)

if orth_init_cont:
    init_fun, predict_fun = stax.serial(
        HomotopyDense(out_dim=18, W_init=orthogonal(), b_init=zeros),
        Dense(out_dim=10, W_init=normal(), b_init=normal())
    )
    #init_fun, predict_fun = Dense(out_dim=10, W_init=normal(), b_init=normal())
else:
    # baseline network
    init_fun, predict_fun = stax.serial(
        Dense(out_dim=18),
        Dense(out_dim=10), LogSoftmax
    )

class ModelContClassifier(AbstractProblem):
    def __init__(self):
        self.HPARAMS_PATH = "hparams.json"

    @staticmethod
    def objective(params, bparam, batch) -> float:
        x, targets = batch
        x = np.reshape(x, (x.shape[0], -1))
        logits = predict_fun(params, x)
        logits = logits - logsumexp(logits, axis=1, keepdims=True)
        loss = -np.mean(np.sum(logits * targets, axis=1))
        loss += 5e-6 * (l2_norm(params)) #+ l2_norm(bparam))
        return loss

    @staticmethod
    def accuracy(params, bparam, batch):
        x, targets = batch
        x = np.reshape(x, (x.shape[0], -1))
        target_class = np.argmax(targets, axis=-1)
        predicted_class = np.argmax(predict_fun(params, x, bparam=bparam[0], activation_func=relu), axis=-1)
        return np.mean(predicted_class == target_class)

    def initial_value(self):
        ae_shape, ae_params = init_fun(random.PRNGKey(0), input_shape)
        bparam = [np.array([0.00], dtype=np.float64)]
        return ae_params, bparam

    def initial_values(self):
        state_0, bparam_0 = self.initial_value()
        state_1 = tree_map(lambda a: a - 0.08, state_0)
        states = [state_0, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam_0)
        bparams = [bparam_0, bparam_1]
        return states, bparams


if __name__ == "__main__":

    problem = ModelContClassifier()
    with open(problem.HPARAMS_PATH, "r") as hfile:
        hparams = json.load(hfile)
    mlflow.set_tracking_uri(hparams['meta']["mlflow_uri"])
    mlflow.set_experiment(hparams['meta']["name"])
    with mlflow.start_run(run_name=hparams['meta']["method"]+"-"+hparams["meta"]["optimizer"]) as run:
        ae_params, bparam = problem.initial_value()
        bparam = pytree_element_add(bparam, 0.0)
        mlflow.log_dict(hparams, artifact_file="hparams/hparams.json")
        artifact_uri = mlflow.get_artifact_uri()
        print("Artifact uri: {}".format(artifact_uri))

        mlflow.log_text("", artifact_file="output/_touch.txt")
        artifact_uri2 = mlflow.get_artifact_uri("output/")
        print("Artifact uri: {}".format(artifact_uri2))
        hparams["meta"]["output_dir"] = artifact_uri2
        file_name = f"{artifact_uri2}/version.jsonl"

        sw = StateWriter(file_name=file_name)

        data_loader = iter(get_mnist_data(batch_size=hparams["batch_size"], resize=True, filter=hparams['filter']))
        num_batches = meta_mnist(batch_size=hparams["batch_size"],filter=hparams['filter'])["num_batches"]
        print(f"num of bathces: {num_batches}")
        compute_grad_fn = jit(grad(problem.objective, [0]))

        opt = OptimizerCreator(hparams["meta"]["optimizer"], learning_rate=hparams["natural_lr"]).get_optimizer()
        ma_loss = []
        for epoch in range(hparams["warmup_period"]):
            for b_j in range(num_batches):
                batch = next(data_loader)
                ae_grads = compute_grad_fn(ae_params, bparam, batch)
                ae_params = opt.update_params(ae_params, ae_grads[0], step_index=epoch)
                #bparam = opt.update_params(bparam, b_grads, step_index=epoch)
                loss = problem.objective(ae_params, bparam, batch)
                ma_loss.append(loss)
                print(f"loss:{loss}  norm:{l2_norm(ae_grads)}")
            opt.lr = exp_decay(epoch, hparams["natural_lr"])
            mlflow.log_metrics({
                "train_loss": float(loss),
                "ma_loss": float(ma_loss[-1]),
                "learning_rate": float(opt.lr),
                "bparam":float(bparam[0]),
                "norm grads": float(l2_norm(ae_grads))
            }, epoch)

            if len(ma_loss) > 100:
                loss_check = running_mean(ma_loss, 50)
                if math.isclose(
                        loss_check[-1], loss_check[-2], abs_tol=hparams["loss_tol"]
                ):
                    print(f"stopping at {epoch}")
                    break

        train_images, train_labels, test_images, test_labels = mnist(
            permute_train=False, resize=True, filter=hparams["filter"]
        )

        val_loss = problem.objective(ae_params, bparam, (test_images, test_labels))
        print(f"val loss: {val_loss, type(ae_params)}")
        val_acc = accuracy(ae_params, bparam, (test_images, test_labels))
        print(f"val acc: {val_acc}")
        mlflow.log_metric("val_acc", float(val_acc))
        mlflow.log_metric("val_loss", float(val_loss))

        q = float(l2_norm(ae_grads[0]))
        if sw:
            sw.write([
                {'u':ae_params},
                {'t': bparam},
                {'f':loss},
                {'q':q},
            ])
        else:
            print('sw none')
    with open(artifact_uri2+'params.pkl', 'wb') as file:
        pickle.dump(ae_params, file)

    with open(artifact_uri2+'params.pkl', 'rb') as file:
        p = pickle.load(file)

    val_loss = problem.objective(p, bparam, (test_images, test_labels))
    print(f"val loss: {val_loss, type(ae_params)}")
    val_acc = accuracy(p, bparam, (test_images, test_labels))
    print(f"val acc: {val_acc}")

    # dg2 = hessian(problem.objective, argnums=[0])(p, bparam, (train_images, train_labels))
    # mtree, _ = ravel_pytree(dg2)
    # eigen = np.linalg.eigvals(mtree.reshape(370, 370)).real
    # eigen = sorted(eigen, reverse=True)
    # print(eigen)
    # print(len(eigen))
    # neg_count = len(list(filter(lambda x: (x < 0), eigen)))
    #
    # # we can also do len(list1) - neg_count
    # pos_count = len(list(filter(lambda x: (x >= 0), eigen)))
    #
    # print("Positive numbers in the eigen: ", pos_count)
    # print("Negative numbers in the eigen: ", neg_count)



from jax.nn.initializers import glorot_uniform, normal
from jax.nn import sigmoid
from jax.experimental.optimizers import l2_norm
from jax.experimental.stax import Sigmoid
from jax import random
import jax.numpy as np
from jax.experimental import stax
import numpy.random as npr
from cjax.utils.custom_nn import HomotopyDense


def generate_data_01():
    batch_size = 8
    input_shape = (batch_size, 4)

    def synth_batches():
        while True:
            images = npr.rand(*input_shape).astype("float32")
            yield images

    batches = synth_batches()
    inputs = next(batches)

    init_func, predict_func = stax.serial(
        HomotopyDense(out_dim=4, W_init=glorot_uniform(), b_init=normal()),
        HomotopyDense(out_dim=1, W_init=glorot_uniform(), b_init=normal()),
        Sigmoid,
    )

    ae_shape, ae_params = init_func(random.PRNGKey(0), input_shape)
    # assert ae_shape == input_shape
    bparam = [np.array([0.0], dtype=np.float64)]
    logits = predict_func(ae_params, inputs, bparam=bparam[0], activation_func=sigmoid)
    loss = np.mean((np.subtract(logits, logits))) + l2_norm(ae_params) + l2_norm(bparam)

    return inputs, logits, ae_params, bparam, init_func, predict_func


if __name__ == "__main__":
    x, y, param, bparam, _, _ = generate_data_01()
    print(x.shape, y.shape)
    print(y)



import numpy.random as npr

from jax.experimental import stax
from jax.experimental.stax import (
    AvgPool,
    BatchNorm,
    Conv,
    Dense,
    FanInSum,
    FanOut,
    Flatten,
    GeneralConv,
    Identity,
    MaxPool,
    Relu,
    LogSoftmax,
)


from cjax.utils.abstract_problem import AbstractProblem
import jax.numpy as np
from jax import random
from jax.experimental.optimizers import l2_norm
from jax.tree_util import *

batch_size = 8
num_classes = 1001
input_shape = (224, 224, 3, batch_size)
step_size = 0.1
num_steps = 10

# ResNet blocks compose other layers


def LambdaIdentity():
    """Layer construction function for an identity layer."""
    init_fun = lambda rng, input_shape: (input_shape, ())
    apply_fun = lambda params, inputs, **kwargs: inputs * kwargs.get("bparams")
    return init_fun, apply_fun


LambdaIdentity = LambdaIdentity()


def ConvBlock(kernel_size, filters, strides=(2, 2)):
    ks = kernel_size
    filters1, filters2, filters3 = filters
    Main = stax.serial(
        Conv(filters1, (1, 1), strides),
        BatchNorm(),
        Relu,
        Conv(filters2, (ks, ks), padding="SAME"),
        BatchNorm(),
        Relu,
        Conv(filters3, (1, 1)),
        BatchNorm(),
    )
    Shortcut = stax.serial(Conv(filters3, (1, 1), strides), BatchNorm())
    return stax.serial(FanOut(2), stax.parallel(Main, Shortcut), FanInSum, Relu)


def IdentityBlock(kernel_size, filters):
    ks = kernel_size
    filters1, filters2 = filters

    def make_main(input_shape):
        # the number of output channels depends on the number of input channels
        return stax.serial(
            Conv(filters1, (1, 1)),
            BatchNorm(),
            Relu,
            Conv(filters2, (ks, ks), padding="SAME"),
            BatchNorm(),
            Relu,
            Conv(input_shape[3], (1, 1)),
            BatchNorm(),
        )

    Main = stax.shape_dependent(make_main)
    return stax.serial(FanOut(2), stax.parallel(Main, Identity), FanInSum, Relu)


# ResNet architectures compose layers and ResNet blocks


def ResNet50(num_classes):
    return stax.serial(
        GeneralConv(("HWCN", "OIHW", "NHWC"), 64, (7, 7), (2, 2), "SAME"),
        BatchNorm(),
        Relu,
        MaxPool((3, 3), strides=(2, 2)),
        ConvBlock(3, [64, 64, 256], strides=(1, 1)),
        IdentityBlock(3, [64, 64]),
        IdentityBlock(3, [64, 64]),
        ConvBlock(3, [128, 128, 512]),
        IdentityBlock(3, [128, 128]),
        IdentityBlock(3, [128, 128]),
        IdentityBlock(3, [128, 128]),
        ConvBlock(3, [256, 256, 1024]),
        IdentityBlock(3, [256, 256]),
        IdentityBlock(3, [256, 256]),
        IdentityBlock(3, [256, 256]),
        IdentityBlock(3, [256, 256]),
        IdentityBlock(3, [256, 256]),
        ConvBlock(3, [512, 512, 2048]),
        IdentityBlock(3, [512, 512]),
        IdentityBlock(3, [512, 512]),
        AvgPool((7, 7)),
        Flatten,
        Dense(num_classes),
        LogSoftmax,
    )


def ResNet(num_classes):
    return stax.serial(
        GeneralConv(("HWCN", "OIHW", "NHWC"), 64, (7, 7), (2, 2), "SAME"),
        BatchNorm(),
        Relu,
        MaxPool((3, 3), strides=(2, 2)),
        ConvBlock(3, [4, 4, 4], strides=(1, 1)),
        IdentityBlock(3, [4, 4]),
        AvgPool((3, 3)),
        Flatten,
        Dense(num_classes),
        LogSoftmax,
    )


def synth_batches():
    rng = npr.RandomState(0)
    while True:
        images = rng.rand(*input_shape).astype("float32")
        labels = rng.randint(num_classes, size=(batch_size, 1))
        onehot_labels = labels == np.arange(num_classes)
        yield images, onehot_labels


batches = synth_batches()
inputs, outputs = next(batches)
init_fun, predict_fun = ResNet(num_classes)


class ResNet50Network(AbstractProblem):
    def __init__(self):
        pass

    @staticmethod
    def objective(params, bparam) -> float:
        logits = predict_fun(params, inputs)
        loss = -jnp.sum(logits * outputs)
        loss += l2_norm(params) + l2_norm(bparam)
        return loss

    def initial_value(self):
        _, init_params = init_fun(random.PRNGKey(0), input_shape)
        bparam = [np.array([-0.50], dtype=np.float64)]
        state = init_params
        return state, bparam

    def initial_values(self):
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.05, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams



from cjax.utils.math_trees import *
from jax.experimental.optimizers import l2_norm
import pickle
from jax.flatten_util import ravel_pytree
from jax.tree_util import tree_flatten
import matplotlib.pyplot as plt

sgd_path = "/opt/ml/mlruns/0/cda8d69d4f1d4abaa3ec34f4294242d7/artifacts/output/params.pkl"
sadam_path = "/opt/ml/mlruns/0/499e2c7d99624bbebe5804832cdae845/artifacts/output/params.pkl"
nadam_path = "/opt/ml/mlruns/0/e62430f647274c6e98d6eab930a0150a/artifacts/output/params.pkl"
ngd_path = "/opt/ml/mlruns/0/236a47c6f2524156963a8cc766a06be4/artifacts/output/param.pkl"

pgd_path = "/opt/ml/mlruns/5/5a518d1762cb4b1eafc73e5cfdb11760/artifacts/output/parc.pkl"
ogd_path = "/opt/ml/mlruns/5/ef3650bc42a945ac8a3b5200bd181283/artifacts/output/params.pkl"
paths = [pgd_path, ogd_path]#, sadam_path, nadam_path]
trees = []
for p in paths:
    with open(p, 'rb') as file:
        tmp = pickle.load(file)
        ent, _ = ravel_pytree(tmp)
        trees.append(ent)

# sgd = trees[0]
# ngd = trees[1]
# sadam = trees[2]
# nadam = trees[3]
# name_trees = ['']

# sgd_ngd = l2_norm(pytree_sub((sgd), (nadam)))
# print("sgd_ngd", sgd_ngd)


#sgd_ngd = l2_norm(pytree_sub(sadam, nadam))
pairs = [(p1, p2) for p1 in trees for p2 in trees]
pairs_dist = []
for k1, k2 in pairs:
    pairs_dist.append(l2_norm(pytree_sub(k1, k2)))
print(pairs_dist)

labels = ['pgd','ogd'] #['sgd', 'ngd','sadam', 'nadam' ]
M = len(labels)
corr_matrix = np.asarray(pairs_dist).reshape(M,M)

fig, ax = plt.subplots()
im = ax.imshow(corr_matrix)
im.set_clim(0, 15)
ax.grid(False)
ax.xaxis.set(ticks=range(M), ticklabels=labels)
ax.yaxis.set(ticks=range(M), ticklabels=labels)
#ax.set_ylim(2.5, -0.5)
for i in range(M):
    for j in range(M):
        ax.text(j, i, corr_matrix[i, j], ha='center', va='center',
                color='r')
cbar = ax.figure.colorbar(im, ax=ax, format='% .2f')
plt.show()
# #pairs_sub = list(map(lambda x,y: pytree_sub(x,y), pairs))
# pairs_norm = list(map(l2_norm, pairs_sub))
# print(pairs_norm)

#### compute distqances


from abc import ABC, abstractmethod
from typing import Tuple


class Corrector(ABC):
    """Abstract Corrector to be inherited by developer for any new corrector."""

    @abstractmethod
    def correction_step(self) -> Tuple:
        pass

    @abstractmethod
    def _assign_states(self):
        pass



from typing import Tuple
from jax import grad, jit
from cjax.continuation.methods.corrector.base_corrector import Corrector


class ConstrainedCorrector(Corrector):
    """Minimize the objective using gradient based method along with some constraint"""

    def __init__(
        self,
        optimizer,
        objective,
        dual_objective,
        lagrange_multiplier,
        concat_states,
        delta_s,
        ascent_opt,
        compute_min_grad_fn,
        compute_max_grad_fn,
        compute_grad_fn,
        hparams,
    ):
        self.concat_states = concat_states
        self._state = None
        self._bparam = None
        self.opt = optimizer
        self.ascent_opt = ascent_opt
        self.objective = objective
        self.dual_objective = dual_objective
        self._lagrange_multiplier = lagrange_multiplier
        self._state_secant_vector = None
        self._state_secant_c2 = None
        self.delta_s = delta_s
        self.warmup_period = hparams["warmup_period"]
        self.ascent_period = hparams["ascent_period"]
        self.descent_period = hparams["descent_period"]
        self.max_norm_state = hparams["max_bounds"]
        self.hparams = hparams
        self.compute_min_grad_fn = compute_min_grad_fn
        self.compute_max_grad_fn = compute_max_grad_fn
        self.compute_grad_fn = compute_grad_fn
        self._assign_states()

    def _assign_states(self):
        self._state = self.concat_states[0]
        self._bparam = self.concat_states[1]
        self._state_secant_vector = self.concat_states[2]
        self._state_secant_c2 = self.concat_states[3]

    def correction_step(self) -> Tuple:
        """Given the current state optimize to the correct state.

        Returns:
          (state: problem parameters, bparam: continuation parameter) Tuple
        """
        for k in range(self.warmup_period):
            grads = self.compute_grad_fn(self._state, self._bparam)
            self._state = self.opt.update_params(self._state, grads[0])

        for k in range(self.ascent_period):
            lagrange_grads = self.compute_max_grad_fn(
                self._state,
                self._bparam,
                self._lagrange_multiplier,
                self._state_secant_c2,
                self._state_secant_vector,
                self.delta_s,
            )
            self._lagrange_multiplier = self.ascent_opt.update_params(
                self._lagrange_multiplier, lagrange_grads[0]
            )
            for j in range(self.descent_period):
                state_grads, bpram_grads = self.compute_min_grad_fn(
                    self._state,
                    self._bparam,
                    self._lagrange_multiplier,
                    self._state_secant_c2,
                    self._state_secant_vector,
                    self.delta_s,
                )
                self._bparam = self.opt.update_params(self._bparam, bpram_grads)
                self._state = self.opt.update_params(self._state, state_grads)

        return self._state, self._bparam



from abc import ABC, abstractmethod
from typing import Dict, Any
import jsonlines
import json
import jax.numpy as np
import glob, os


class StateVariable:
    """To track the state of problem"""

    def __init__(self, state: Any, counter: int):
        self._state = state
        self._counter = counter

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        self._state = state

    @property
    def counter(self):
        return self._counter

    @counter.setter
    def counter(self, counter):
        self._counter = counter

    def get_record(self) -> Dict:
        """Get the state as indexed by the continuation counter"""
        return {self._counter: self._state}


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class StateWriter:
    """State Writer will write the state values to a file."""

    def __init__(self, file_name: str):
        """Create a file object"""
        self.file_name = file_name
        self._clear()
        self.writer = jsonlines.Writer(
            fp=open(file_name, mode="a", encoding="utf-8"), dumps=NumpyEncoder().encode, flush=False
        )

    def _clear(self):
        try:
            for f in glob.glob(self.file_name):
                print(f"Clearing previous runs of {f}")
                os.remove(f)
        except Exception as e:
            print(f"No previos runs. {e}")

    def write(self, record: list):
        """Write/Append the record to the file."""
        self.writer.write(record)



import jax.numpy as np
import numpy as onp
from jax.experimental import stax
from jax.nn.initializers import zeros, orthogonal, delta_orthogonal, uniform
from jax.nn import sigmoid, hard_tanh, soft_sign, relu
from jax.experimental.stax import Dense, Sigmoid, Softplus, Tanh, Relu
from cjax.optimizer.optimizer import GDOptimizer, AdamOptimizer, OptimizerCreator
import numpy.random as npr
from jax import random
from jax import jit, vmap, grad
from cjax.utils.abstract_problem import AbstractProblem
from jax.tree_util import tree_map
from cjax.utils.math_trees import *
from cjax.utils.evolve_utils import *
from cjax.continuation.states.state_variables import StateWriter
import math
import mlflow
import json
from cjax.utils.custom_nn import constant_2d, HomotopyDense, v_2d
from cjax.utils.datasets import mnist, get_mnist_data, meta_mnist, mnist_preprocess_cont, get_mnist_batch_alter
import pickle
from examples.torch_data import get_data

data_size = 50000
input_shape = (data_size, 36)
step_size = 0.1
num_steps = 10
code_dim = 1
npr.seed(7)

with open("hparams.json", "r") as hfile:
    t_hparams = json.load(hfile)

if t_hparams['init_m'] == 'pca':
    train_images, labels, _, _ = mnist_preprocess_cont(resize=True, filter=False)
    inputs = train_images
    inputs = center_data(inputs)
    u, s, v_t = onp.linalg.svd(inputs, full_matrices=False)
    I = np.eye(v_t.shape[-1])
    I_add = npr.normal(0.0, 0.002, size=I.shape)
    noisy_I = I + I_add

    init_fun, predict_fun = stax.serial(
        HomotopyDense(out_dim=4, W_init=v_2d(v_t), b_init=zeros),
        HomotopyDense(out_dim=2, W_init=constant_2d(noisy_I), b_init=zeros),
        HomotopyDense(out_dim=4, W_init=constant_2d(noisy_I), b_init=zeros),
        Dense(out_dim=input_shape[-1], W_init=v_2d(v_t.T), b_init=zeros),
    )
    del inputs, train_images, _
elif t_hparams['init_m'] == 'orth':
        init_fun, predict_fun = stax.serial(
        HomotopyDense(out_dim=4, W_init=orthogonal(), b_init=zeros),
        HomotopyDense(out_dim=2, W_init=orthogonal(), b_init=zeros),
        HomotopyDense(out_dim=4, W_init=orthogonal(), b_init=zeros),
        Dense(out_dim=input_shape[-1], W_init=orthogonal(), b_init=zeros),
    )
else:
    if t_hparams['hidden_activation'] == 'relu':
        init_fun, predict_fun = stax.serial(
            Dense(out_dim=4), Relu,
            Dense(out_dim=2), Relu,
            Dense(out_dim=4), Relu,
            Dense(out_dim=input_shape[-1]),
        )
    elif t_hparams['hidden_activation'] == 'sigmoid':
        init_fun, predict_fun = stax.serial(
            Dense(out_dim=4), Sigmoid,
            Dense(out_dim=2), Sigmoid,
            Dense(out_dim=4), Sigmoid,
            Dense(out_dim=input_shape[-1]),
        )
    else:
        init_fun, predict_fun = stax.serial(
            Dense(out_dim=4),
            Dense(out_dim=2),
            Dense(out_dim=4),
            Dense(out_dim=input_shape[-1]),
        )


class PCATopologyAE(AbstractProblem):
    def __init__(self):
        self.HPARAMS_PATH = "hparams.json"

    @staticmethod
    def objective(params, bparam, batch) -> float:
        x, _ = batch
        x = np.reshape(x, (x.shape[0], -1))
        logits = predict_fun(params, x, bparam=bparam[0], activation_func=sigmoid)
        loss = np.mean(np.square((np.subtract(logits, x))))
        #loss += 5e-6 * (l2_norm(params)) #+ l2_norm(bparam))
        return loss

    @staticmethod
    def accuracy(params, bparam, batch) -> float:
        x, _ = batch
        x = np.reshape(x, (x.shape[0], -1))
        logits = predict_fun(params, x, bparam=bparam[0], activation_func=sigmoid)
        loss = np.mean(np.square((np.subtract(logits, x))))
        #loss += 5e-6 * (l2_norm(params))  # + l2_norm(bparam))
        return loss

    def initial_value(self):
        ae_shape, ae_params = init_fun(random.PRNGKey(0), input_shape)
        # print(ae_shape)
        assert ae_shape == input_shape
        bparam = [np.array([0.00], dtype=np.float64)]
        return ae_params, bparam

    def initial_values(self):
        state_0, bparam_0 = self.initial_value()
        state_1 = tree_map(lambda a: a - 0.08, state_0)
        states = [state_0, state_1]
        bparam_1 = tree_map(lambda a: a + 0.05, bparam_0)
        bparams = [bparam_0, bparam_1]
        return states, bparams



if __name__ == "__main__":

    problem = PCATopologyAE()
    with open(problem.HPARAMS_PATH, "r") as hfile:
        hparams = json.load(hfile)

    train_images, train_labels, test_images, test_labels = mnist_preprocess_cont(
        resize=True, filter=hparams["filter"]
    )
    mlflow.set_tracking_uri(hparams['meta']["mlflow_uri"])
    mlflow.set_experiment(hparams['meta']["name"])
    with mlflow.start_run(run_name=hparams['meta']["method"]+"-"+hparams["meta"]["optimizer"]) as run:
        ae_params, bparam = problem.initial_value()
        bparam = pytree_element_add(bparam, 0.0)
        mlflow.log_dict(hparams, artifact_file="hparams/hparams.json")
        artifact_uri = mlflow.get_artifact_uri()
        print("Artifact uri: {}".format(artifact_uri))

        mlflow.log_text("", artifact_file="output/_touch.txt")
        artifact_uri2 = mlflow.get_artifact_uri("output/")
        print("Artifact uri: {}".format(artifact_uri2))
        hparams["meta"]["output_dir"] = artifact_uri2
        file_name = f"{artifact_uri2}/version.jsonl"

        sw = StateWriter(file_name=file_name)
        data_loader = iter(get_mnist_batch_alter(train_images, train_labels, test_images, test_labels,
                                                 [0.0], batch_size=hparams["batch_size"],
                                                 resize=True, filter=hparams['filter'] ))
        #data_loader = iter(get_mnist_data(batch_size=hparams["batch_size"], resize=True, filter=hparams['filter']))
        num_batches = meta_mnist(batch_size=hparams["batch_size"],filter=hparams['filter'])["num_batches"]
        print(f"num of bathces: {num_batches}")
        compute_grad_fn = jit(grad(problem.objective, [0]))

        opt = OptimizerCreator(hparams["meta"]["optimizer"], learning_rate=hparams["natural_lr"]).get_optimizer()
        ma_loss = []
        for epoch in range(500):
            for b_j in range(num_batches):
                batch = next(data_loader)
                grads = compute_grad_fn(ae_params, bparam, batch)
                ae_params = opt.update_params(ae_params, grads[0], step_index=epoch)
                loss = problem.objective(ae_params, bparam, batch)
                ma_loss.append(loss)
                print(f"loss:{loss}  norm:{l2_norm(grads)}")
            opt.lr = exp_decay(epoch, hparams["descent_lr"])
            mlflow.log_metrics({
                "train_loss": float(loss),
                "ma_loss": float(ma_loss[-1]),
                "learning_rate": float(opt.lr),
                "norm grads": float(l2_norm(grads))
            }, epoch)

            if len(ma_loss) > 100:
                loss_check = running_mean(ma_loss, 50)
                if math.isclose(
                        loss_check[-1], loss_check[-2], abs_tol=hparams["loss_tol"]
                ):
                    print(f"stopping at {epoch}")
                    break



        val_loss = problem.objective(ae_params, bparam, (test_images, test_labels))
        print(f"val loss: {val_loss, type(ae_params)}")

        mlflow.log_metric("val_loss", float(val_loss))


        q = float(l2_norm(grads[0]))
        if sw:
            sw.write([
                {'u':ae_params},
                {'t': bparam},
                {'f':loss},
                {'q':q},
            ])
        else:
            print('sw none')
    with open(artifact_uri2+'params.pkl', 'wb') as file:
        pickle.dump(ae_params, file)



"""
Inspired from:
@misc{gehring2019fax,
  author = {Clement Gehring, Pierre-Luc Bacon, Florian Schaefer},
  title = {{FAX: differentiating fixed point problems in JAX}},
  note = {Available at: https://github.com/gehring/fax},
  year = {2019}
}
"""

import operator

from jax import lax
from jax import numpy as np
from jax import tree_util, flatten_util
from jax.experimental.optimizers import l2_norm


def pytree_dot(x, y) -> float:
    partial_dot = tree_util.tree_multimap(lambda arr1, arr2: np.sum(arr1 * arr2), x, y)
    return tree_util.tree_reduce(lax.add, partial_dot)


def pytree_sub(x, y):
    return tree_util.tree_multimap(lax.sub, x, y)


def pytree_array_equal(x, y):
    is_eq = tree_util.tree_multimap(lambda arr1, arr2: np.array_equal(arr1, arr2), x, y)
    return tree_util.tree_reduce(operator.and_, is_eq)


def pytree_shape_array_equal(x, y):
    is_eq = tree_util.tree_multimap(lambda arr1, arr2: (arr1.shape == arr2.shape), x, y)
    return tree_util.tree_reduce(operator.and_, is_eq)


def pytree_zeros_like(x):
    return tree_util.tree_map(lambda arr: 0 * arr, x)


def pytree_ones_like(x):
    return tree_util.tree_map(lambda arr: 0 * arr + 1, x)


def pytree_element_add(x, s):
    return tree_util.tree_map(lambda a: a + s, x)


def pytree_element_mul(x, s):
    return tree_util.tree_map(lambda a: a * s, x)


def pytree_to_vec(x):  # cannot be jitted as it return pytree , unravel function
    return flatten_util.ravel_pytree(x)


def pytree_normalized(x):
    return tree_util.tree_map(lambda a: a / l2_norm(x), x)


def pytree_relative_error(x, y):
    partial_error = tree_util.tree_multimap(
        lambda a, b: l2_norm(pytree_sub(a, b)) / (l2_norm(a) + 1e-5), x, y
    )
    return tree_util.tree_reduce(lax.add, partial_error)



from cjax.continuation._arc_len_continuation import Continuation
from cjax.continuation.states.state_variables import StateWriter
from cjax.utils.data_img_gamma import mnist_gamma
from cjax.utils.datasets import mnist
from cjax.continuation.methods.predictor.arc_secant_predictor import SecantPredictor
from cjax.continuation.methods.corrector.perturb_parc_evolve import (
    PerturbedFixedCorrecter,
)
from cjax.continuation.methods.corrector.unconstrained_corrector import (
    UnconstrainedCorrector,
)
import jax.numpy as np
from jax.tree_util import *
import copy
from cjax.utils.profiler import profile
from jax.experimental.optimizers import l2_norm
from cjax.continuation.states.state_variables import StateVariable, StateWriter
from jax import jit, grad
import numpy.random as npr
from cjax.utils.math_trees import pytree_element_add
import mlflow
# TODO: make **kwargs availible


class PerturbedPseudoArcLenFixedContinuation(Continuation):
    """Noisy Pseudo Arc-length Continuation strategy.

    Composed of secant predictor and noisy constrained corrector"""

    def __init__(
        self,
        state,
        bparam,
        state_0,
        bparam_0,
        counter,
        objective,
        dual_objective,
        accuracy_fn,
        hparams,
        key_state,
    ):

        # states
        self._state_wrap = StateVariable(state, counter)
        self._bparam_wrap = StateVariable(
            bparam, counter
        )  # Todo : save tree def, always unlfatten before compute_grads
        self._prev_state = state_0
        self._prev_bparam = bparam_0

        # objectives
        self.objective = objective
        self.dual_objective = dual_objective
        self.accuracy_fn1= jit(accuracy_fn)
        self.value_func = jit(self.objective)

        self.hparams = hparams
        if hparams["meta"]["dataset"] == "mnist":
            if hparams["continuation_config"] == 'data':
                self.dataset_tuple = mnist_gamma(
                    resize=hparams["resize_to_small"],
                    filter=hparams["filter"])
            else:
                self.dataset_tuple = mnist(
                    resize=hparams["resize_to_small"],
                    filter=hparams["filter"])

        self._value_wrap = StateVariable(
            0.06, counter
        )  # TODO: fix with a static batch (test/train)
        self._quality_wrap = StateVariable(
            l2_norm(self._state_wrap.state) / 10, counter
        )

        # every step hparams
        self.continuation_steps = hparams["continuation_steps"]

        self._delta_s = hparams["delta_s"]
        self._prev_delta_s = hparams["delta_s"]
        self._omega = hparams["omega"]

        # grad functions # should be pure functional
        self.compute_min_grad_fn = jit(grad(self.dual_objective, [0, 1]))
        self.compute_grad_fn = jit(grad(self.objective, [0]))

        # extras
        self.state_tree_def = None
        self.bparam_tree_def = None
        self.output_file = hparams["meta"]["output_dir"]
        self.prev_secant_direction = None
        self.perturb_index = key_state
        self.sw = StateWriter(f"{self.output_file}/version_{self.perturb_index}.json")
        self.key_state = key_state + npr.randint(100, 200)
        self.clip_lambda_max = lambda g: np.where(
            (g > self.hparams["lambda_max"]), self.hparams["lambda_max"], g
        )
        self.clip_lambda_min = lambda g: np.where(
            (g < self.hparams["lambda_min"]), self.hparams["lambda_min"], g
        )

    @profile(sort_by="cumulative", lines_to_print=10, strip_dirs=True)
    def run(self):
        """Runs the continuation strategy.

        A continuation strategy that defines how predictor and corrector components of the algorithm
        interact with the states of the mathematical system.
        """
        for i in range(self.continuation_steps):
            self._state_wrap.counter = i
            self._bparam_wrap.counter = i
            self._value_wrap.counter = i
            self._quality_wrap.counter = i

            if i == 0 and self.hparams["natural_start"]:
                print(f" unconstrained solver for 1st step")
                concat_states = [
                    self._prev_state,
                    pytree_element_add(self._prev_bparam, 0.03),
                ]

                corrector = UnconstrainedCorrector(
                    objective=self.objective,
                    concat_states=concat_states,
                    grad_fn=self.compute_grad_fn,
                    value_fn=self.value_func,
                    accuracy_fn=self.accuracy_fn1,
                    hparams=self.hparams,
                    dataset_tuple=self.dataset_tuple,
                )
                state, bparam, quality, value, val_loss, val_acc = corrector.correction_step()
                if self.hparams["double_natural_start"]:  # TODO: refactor natural and double natural start
                    self._prev_state = state
                    self._prev_bparam = bparam
                    print(f" unconstrained solver for 2nd step")
                    concat_states = [
                        self._prev_state,
                        pytree_element_add(self._prev_bparam, 0.07),
                    ]

                    corrector = UnconstrainedCorrector(
                        objective=self.objective,
                        concat_states=concat_states,
                        grad_fn=self.compute_grad_fn,
                        value_fn=self.value_func,
                        accuracy_fn=self.accuracy_fn1,
                        hparams=self.hparams,
                        dataset_tuple=self.dataset_tuple,
                    )
                    state, bparam, quality, value, val_loss, val_acc = corrector.correction_step()

                self._state_wrap.state = state
                self._bparam_wrap.state = bparam

            print("delta_s",
                self._value_wrap.get_record(),
                self._bparam_wrap.get_record(),
                self._delta_s,
            )
            self.sw.write(
                [
                    self._state_wrap.get_record(),
                    self._bparam_wrap.get_record(),
                    self._value_wrap.get_record(),
                    self._quality_wrap.get_record(),
                ]
            )

            concat_states = [
                (self._prev_state, self._prev_bparam),
                (self._state_wrap.state, self._bparam_wrap.state),
                self.prev_secant_direction,
            ]

            predictor = SecantPredictor(
                concat_states=concat_states,
                delta_s=self._delta_s,
                prev_delta_s=self._prev_delta_s,
                omega=self._omega,
                net_spacing_param=self.hparams["net_spacing_param"],
                net_spacing_bparam=self.hparams["net_spacing_bparam"],
                hparams=self.hparams,
            )
            predictor.prediction_step()

            self.prev_secant_direction = predictor.secant_direction

            self.hparams["sphere_radius"] = (
                self.hparams["sphere_radius_m"] * self._delta_s
            )  # l2_norm(predictor.secant_direction)
            mlflow.log_metric(f"sphere_radius{self.perturb_index}", self.hparams["sphere_radius"], i)
            mlflow.log_metric(f"delta_s{self.perturb_index}", self._delta_s, i)
            concat_states = [
                predictor.state,
                predictor.bparam,
                predictor.secant_direction,
                {"state": predictor.state, "bparam": predictor.bparam},
            ]
            corrector = PerturbedFixedCorrecter(
                objective=self.objective,
                dual_objective=self.dual_objective,
                accuracy_fn1=self.accuracy_fn1,
                value_fn=self.value_func,
                concat_states=concat_states,
                key_state=self.key_state,
                compute_min_grad_fn=self.compute_min_grad_fn,
                compute_grad_fn=self.compute_grad_fn,
                hparams=self.hparams,
                delta_s=self._delta_s,
                pred_state=[self._state_wrap.state, self._bparam_wrap.state],
                pred_prev_state=[self._state_wrap.state, self._bparam_wrap.state],
                counter=self.continuation_steps,
                dataset_tuple=self.dataset_tuple,
            )
            self._prev_state = copy.deepcopy(self._state_wrap.state)
            self._prev_bparam = copy.deepcopy(self._bparam_wrap.state)

            (
                state,
                bparam,
                quality,
                value,
                val_loss,
                val_acc,
                corrector_omega,
            ) = (
                corrector.correction_step()
            )  # TODO: make predictor corrector similar api's
            # TODO: Enable MLFlow
            bparam = tree_map(self.clip_lambda_max, bparam)
            bparam = tree_map(self.clip_lambda_min, bparam)

            self._state_wrap.state = state
            self._bparam_wrap.state = bparam
            self._value_wrap.state = value
            self._quality_wrap.state = quality
            # self._omega = corrector_omega
            self._prev_delta_s = self._delta_s
            self._delta_s = corrector_omega * self._delta_s
            self._delta_s = min(self._delta_s, self.hparams["max_arc_len"])
            self._delta_s = max(self._delta_s, self.hparams["min_arc_len"])

            if (bparam[0] >= self.hparams["lambda_max"]) or (
                bparam[0] <= self.hparams["lambda_min"]
            ):
                self.sw.write(
                    [
                        self._state_wrap.get_record(),
                        self._bparam_wrap.get_record(),
                        self._value_wrap.get_record(),
                        self._quality_wrap.get_record(),
                    ]
                )
                break
            mlflow.log_metrics({
                f"train_loss{self.perturb_index}": float(self._value_wrap.state),
                f"delta_s{self.perturb_index}": float(self._delta_s),
                f"norm grads{self.perturb_index}": float(self._quality_wrap.state),
                f"val_loss{self.perturb_index}": float(val_loss),
                f"val_acc{self.perturb_index}": float(val_acc),
                f"corrector_omega{self.perturb_index}": float(corrector_omega)
            }, i)



from cjax.continuation.base_continuation import Continuation
from cjax.continuation.states.state_variables import StateVariable, StateWriter
from cjax.continuation.methods.predictor.secant_predictor import SecantPredictor
from cjax.continuation.methods.corrector.constrained_corrector import (
    ConstrainedCorrector,
)
from jax import jit, grad
import gc
from cjax.utils.profiler import profile
from cjax.optimizer.optimizer import OptimizerCreator
from jax.experimental.optimizers import l2_norm


class PseudoArcLenContinuation(Continuation):
    # May be refactor to only one continuation TODO
    """Pseudo Arc-length Continuation strategy.

    Composed of secant predictor and constrained corrector"""

    def __init__(
        self,
        state,
        bparam,
        state_0,
        bparam_0,
        counter,
        objective,
        dual_objective,
        hparams,
    ):
        # states
        self._state_wrap = StateVariable(state, counter)
        self._bparam_wrap = StateVariable(
            bparam, counter
        )  # Todo : save tree def, always unlfatten before compute_grads
        self._prev_state = state_0
        self._prev_bparam = bparam_0

        # objectives
        self.objective = objective
        self.dual_objective = dual_objective
        self.value_func = jit(self.objective)

        self.hparams = hparams

        self._value_wrap = StateVariable(
            1.0, counter
        )  # TODO: fix with a static batch (test/train)
        self._quality_wrap = StateVariable(l2_norm(self._state_wrap.state), counter)

        # optimizer
        self.opt = OptimizerCreator(
            opt_string=hparams["meta"]["optimizer"], learning_rate=hparams["natural_lr"]
        ).get_optimizer()
        self.ascent_opt = OptimizerCreator(
            opt_string=hparams["meta"]["ascent_optimizer"],
            learning_rate=hparams["ascent_lr"],
        ).get_optimizer()

        # every step hparams
        self.continuation_steps = hparams["continuation_steps"]
        self._lagrange_multiplier = hparams["lagrange_init"]

        self._delta_s = hparams["delta_s"]
        self._omega = hparams["omega"]

        # grad functions # should be pure functional
        self.compute_min_grad_fn = jit(grad(self.dual_objective, [0, 1]))
        self.compute_max_grad_fn = jit(grad(self.dual_objective, [2]))
        self.compute_grad_fn = jit(grad(self.objective, [0]))

        # extras
        self.sw = None
        self.state_tree_def = None
        self.bparam_tree_def = None
        self.output_file = hparams["meta"]["output_dir"]
        self.prev_secant_direction = None

    @profile(sort_by="cumulative", lines_to_print=10, strip_dirs=True)
    def run(self):
        """Runs the continuation strategy.

        A continuation strategy that defines how predictor and corrector components of the algorithm
        interact with the states of the mathematical system.
        """
        self.sw = StateWriter(f"{self.output_file}/version.json")
        for i in range(self.continuation_steps):

            self._state_wrap.counter = i
            self._bparam_wrap.counter = i
            self._value_wrap.counter = i
            self.sw.write(
                [
                    self._state_wrap.get_record(),
                    self._bparam_wrap.get_record(),
                    self._value_wrap.get_record(),
                ]
            )

            concat_states = [
                (self._state_wrap.state, self._bparam_wrap.state),
                (self._prev_state, self._prev_bparam),
                self.prev_secant_direction,
            ]

            predictor = SecantPredictor(
                concat_states=concat_states,
                delta_s=self._delta_s,
                omega=self._omega,
                net_spacing_param=self.hparams["net_spacing_param"],
                net_spacing_bparam=self.hparams["net_spacing_bparam"],
                hparams=self.hparams,
            )
            predictor.prediction_step()
            self.prev_secant_direction = predictor.secant_direction
            self._prev_state = self._state_wrap.state
            self._prev_bparam = self._bparam_wrap.state

            concat_states = [
                predictor.state,
                predictor.bparam,
                predictor.secant_direction,
                predictor.get_secant_concat(),
            ]
            del predictor
            gc.collect()
            corrector = ConstrainedCorrector(
                optimizer=self.opt,
                objective=self.objective,
                dual_objective=self.dual_objective,
                lagrange_multiplier=self._lagrange_multiplier,
                concat_states=concat_states,
                delta_s=self._delta_s,
                ascent_opt=self.ascent_opt,
                compute_min_grad_fn=self.compute_min_grad_fn,
                compute_max_grad_fn=self.compute_max_grad_fn,
                compute_grad_fn=self.compute_grad_fn,
                hparams=self.hparams,
            )
            state, bparam = corrector.correction_step()
            value = self.value_func(state, bparam)
            self._state_wrap.state = state
            self._bparam_wrap.state = bparam
            self._value_wrap.state = value
            del corrector
            gc.collect()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

install_requires = [
    "sphinx_rtd_theme==0.5.1",
    "Sphinx==3.4.3",
    "matplotlib==3.1.2",
    "flax==0.3.0",
    "numpy==1.18.1",
    "jsonlines==2.0.0",
    "jax==0.2.9",
    "jaxlib",
    "nbsphinx",
]


setuptools.setup(
    name="continuation_jax",  # Replace with your own username
    version="0.0.3",
    author="Harsh Nilesh Pathak",
    author_email="harshnpathak@gmail.com",
    description="Continuation Methods for Deep Neural Networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/harsh306/continuation-jax",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.6",
)



import jax.numpy as np
from jax.experimental import stax
from jax.nn.initializers import zeros, orthogonal
from jax.experimental.stax import Dense, Sigmoid
import numpy.random as npr
from jax import random
from cjax.utils.abstract_problem import AbstractProblem
from jax.tree_util import tree_map
from cjax.utils.custom_nn import HomotopyDropout
from cjax.utils.datasets import mnist

data_size = 40000
input_shape = (data_size, 36)
npr.seed(7)


def synth_batches():
    while True:
        images = npr.rand(*input_shape).astype("float32")
        yield images


# batches = synth_batches()
# inputs = next(batches)

# train_images, labels, _, _ = mnist(permute_train=True, resize=True)
# del _
# inputs = train_images[:data_size]
#
# del train_images

# u, s, v_t = onp.linalg.svd(inputs, full_matrices=False)
# I = np.eye(v_t.shape[-1])
# I_add = npr.normal(0.0, 0.002, size=I.shape)
# noisy_I = I + I_add


init_fun, predict_fun = stax.serial(
    HomotopyDropout(rate=0.0),
    # Dense(4, b_init=zeros),
    Dense(4, W_init=orthogonal(scale=1.0), b_init=zeros),
    Sigmoid,
    # Dense(4, b_init=zeros),
    Dense(out_dim=input_shape[-1], W_init=orthogonal(scale=1.0), b_init=zeros),
)
_, key = random.split(random.PRNGKey(0))


class DataTopologyAE(AbstractProblem):
    def __init__(self):
        self.HPARAMS_PATH = "hparams.json"

    @staticmethod
    def objective(params, bparam, batch) -> float:
        x, _ = batch
        x = np.reshape(x, (x.shape[0], -1))
        logits = predict_fun(params, x, bparam=bparam[0], rng=key)
        keep = random.bernoulli(key, bparam[0], x.shape)
        inputs_d = np.where(keep, x, 0)

        loss = np.mean(np.square((np.subtract(logits, inputs_d))))
        # loss += 0.1*(l2_norm(params) + l2_norm(bparam))
        return loss

    def initial_value(self):
        ae_shape, ae_params = init_fun(random.PRNGKey(0), input_shape)
        assert ae_shape == input_shape
        bparam = [np.array([0.01], dtype=np.float64)]
        return ae_params, bparam

    def initial_values(self):
        state, bparam = self.initial_value()
        state_1 = tree_map(lambda a: a + 0.084, state)
        states = [state, state_1]
        bparam_1 = tree_map(lambda a: a + 0.08, bparam)
        bparams = [bparam, bparam_1]
        return states, bparams


if __name__ == "__main__":
    problem = DataTopologyAE()
    ae_params, bparams = problem.initial_value()

    loss = problem.objective(ae_params, bparams)

    print(loss)

    # init_c = constant_2d(I)
    # print(init_c(key=0, shape=(8,8)))



from unittest import TestCase
import jax.numpy as np
import json
from cjax.continuation.methods.predictor.arc_secant_predictor import SecantPredictor

class TestSecantPredictor(TestCase):
    def setUp(self) -> None:
        self.x = [np.array([0.5, 0.5]), 0.5, {"beta": 0.5}]
        self.y = [np.array([0.5, 0.5]), 0.5, {"beta": 0.5}]
        self.z = [np.array([0.0, 0.0]), 0.0, {"beta": 0.0}]
        self.x1 = [np.array([1.0, 1.0]), 1.0, {"beta": 1.0}]
        self.bparam = [np.array([0.5])]
        concat_states = [
            (self.x, self.bparam),
            (self.z, self.bparam),
            None,
        ]
        with open("../../../../examples/autoencoder/hparams.json", 'r') as f:
            hparams = json.load(f)
        self.predictor  = SecantPredictor(concat_states, delta_s=0.02,
                                          prev_delta_s=0.03,omega=1.0,
                                          net_spacing_param=0.1,
                                          net_spacing_bparam=0.2, hparams=hparams)


class TestPredictionStep(TestSecantPredictor):
    def test_prediction_step(self):
        self.predictor._assign_states()
        print(self.predictor.state)
        self.predictor.prediction_step()
        print(self.predictor.state, self.predictor.bparam, self.predictor.hparams["delta_s"])
        self.assertEqual(1,1)
